import builtins
import operator
import itertools

from copy import deepcopy

from collections.abc import Iterable

from typing import Union
from typing import TypeVar
from typing import Type
from typing import Dict 
from typing import List

from abc import ABC
from abc import abstractmethod

from functools import wraps
from functools import reduce

from numpy import linspace
from numpy import prod
from numpy import unravel_index

from .exceptions import Error

import warnings


def custom_formatwarning(msg, *args, **kwargs):
    return "Warning: " + str(msg) + '\n'


warnings.formatwarning = custom_formatwarning


T = TypeVar("T")

StringDict = Dict[str, str]
StringList = List[str]


class Settings:
    """A base class for building python objects out of :obj:`dict` object.

    This class is designed as a recursive method of building python objects
    from json files.

    If the parsed json file does not match the defined settings class
    in attribute names and type, it will catch it an terminate with an error
    message.

    Example
    -------

    The following is a simple use case::

        class UserSettings(Settings):

            @Settings.assign
            def __init__(self, values: dict):
                self.setting_1 = int
                self.setting_2 = SubSetting

        class SubSetting(Settings):

            @Settings.assign
            def __init__(self, values: dict):
                self.subsetting_1 = str

    The `values` parameter should be, for example::

        values = {
            "setting_1": 1,
            "setting_2": {
                subsetting_1: "this_is_a_string"
            }
        }

    This results in::

        settings = UserSettings(values)

        print(settings.setting_1)
            # 1

        print(settings.setting_2.subsetting_1)
            # this_is_a_string

    """
    @staticmethod
    def assign(method):
        """A decorator that applies the
        :meth:`~.Settings.distribute` to parameter of the
        constructor of the derived class.

        """
        @wraps(method)
        def wrapper(self, *args):
            method(self, *args)
            self.distribute(*args)
            try:
                self.consistency_check()
            except AttributeError:
                pass
        return wrapper

    @property
    def primitive(self):
        """:obj:`list`(:obj:`type`) : a list of built in types.

        """
        return [getattr(builtins, d) for d in dir(builtins) if
                isinstance(getattr(builtins, d), type)]

    def distribute(self, values: dict):
        """The method which loops over the attribute/type pairs in the derived
        class' constructor and tries to find and assign values to them from the
        `values` parameter.

        Parameters
        ----------
        values : :obj:`dict`
            The dictionary that contains the values to be assigned to the
            derived class' attributes.

        Returns
        -------
        None

        Raises
        ------
        :class:`~.SettingNotFoundError`
            If one of the derived class' attributes cannot be found in
            `values`.

        :class:`~.SettingTypeError`
            If one of the type of one of the settings found in `values`
            is not of the required type, as defined in the derived class'
            constructor.
        
        :class:`~.SettingCheckError`
            If a check error is caught when assigning a setting to a
            :class:`~.Terminus` instance.

        Note
        ----
        If the setting found in the `values` is None, then this function will
        assign None as the setting value by default, independent of the expected
        value.

        """
        for setting, setting_type in self.__dict__.items():
            if not isinstance(values, dict):
                raise SettingTypeError(dict, type(values))
            try:
                try:
                    value = values[setting]
                except KeyError:
                    raise SettingNotFoundError()
            except SettingNotFoundError as e:
                raise SettingErrorMessage(setting, original_error=e)
            if value is None:
                setattr(self, setting, value)
            elif setting_type not in self.primitive:
                try:
                    setattr(self, setting, setting_type(value))
                except SettingTypeError as e:
                    raise SettingErrorMessage(setting, original_error=e)
                except SettingCheckError as e:
                    raise SettingErrorMessage(setting, original_error=e)
                except SettingNotFoundError as e:
                    raise SettingErrorMessage(setting, original_error=e)
                except SettingRangeTypeError as e:
                    raise SettingErrorMessage(setting, original_error=e)
                except SettingRangeKeyError as e:
                    raise SettingErrorMessage(setting, original_error=e)
                except SettingStringSelectionError as e:
                    raise SettingErrorMessage(setting, original_error=e)
                except ConsistencyError as e:
                    raise SettingErrorMessage(setting, original_error=e)
                except SettingErrorMessage as e:
                    raise SettingErrorMessage(setting, branch_error=e)
            else:
                try:
                    if isinstance(value, setting_type):
                        setattr(self, setting, value)
                    else:
                        raise SettingTypeError(setting_type, type(setting))
                except SettingErrorMessage as e:
                    SettingErrorMessage(setting, original_error=e)
        self.__source__ = values

    def __getattribute__(self, name):
        rv = object.__getattribute__(self, name)
        if issubclass(type(rv), Terminus):
            return rv.get
        elif issubclass(type(rv), Dict):
            return rv.get
        elif issubclass(type(rv), List):
            return rv.get
        elif issubclass(type(rv), Number):
            return rv.get
        else:
            return rv

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __hash__(self):
        return hash(str(self.__source__))
    
    def __eq__(self, other):
        if other.__hash__() == self.__hash__():
            return True
        else:
            return False


class Terminus(ABC):
    """A base class for a terminating setting.

    This class is used to store a single setting in a :class:`~.Settings`
    derived class.

    If the expected type of a :class:`~.Settings` derived class attribute
    inherits from :class:`Termninus`, then when __getattribute__ is called to
    access the :class:`Terminus` instance, it will return the :attr:`value`
    attribute instead.

    This structure allows for value checks to be applied to individual settings,
    on instantiation of :class:`~.Settings` derived classes.

    As with the :class:`~.Settings` derived classes, the decorator
    :meth:`Terminus.assign` should be applied to :class:`Terminus` derived class
    constructors. If this is done, the abstract method :meth:`Terminus.check` is
    automatically called after the derived class' constructor.

    Attributes
    ----------
    type : :obj:`type`
        The attribute defined in the child class that determines the type of the
        setting stored in the class.

    value : :obj:`Any`
        The value stored after all checks have been passed.

    """
    @property
    def get(self):
        return self.value

    @staticmethod
    def assign(method):
        """A decorator that applies the :meth:`Terminus.distribute` to
        parameter of the constructor of the derived class. It will also then
        call the :meth:`Terminus.check` function.

        """
        @wraps(method)
        def wrapper(self, *args):
            method(self, *args)
            self.distribute(*args)
            self.action()
        return wrapper

    @abstractmethod
    def check(self):
        """An abstract method that defines the value checks to be performed on
        the settings stored in this class.

        Raises
        ------
        :class:`~.TypeError`
            If the passed value is of the wrong type.
        
        :obj:`ValueError`
            If the passed value is outside acceptable bounds.

        """
        pass

    def action(self):
        """Method that handles exceptions thrown by the :meth:`Terminus.check`
        function.

        Parameters
        ----------
        check_method
            The abstract check method implementation to be called.

        Raises
        ------
        :class:`~.SettingCheckError`
            If the check function raises any exceptions.

        """
        try:
            self.check()
        except TypeError as e:
            raise SettingCheckError(e, self.value)
        except ValueError as e:
            raise SettingCheckError(e, self.value)

    def distribute(self, value: T):
        """Method called by the decorator :meth:`Terminus.assign` that
        tries to assign the value passed to the constructor of the
        :class:`Terminus` derived class.

        Parameters
        ----------
        value : :class:`TypeVar`
            The value passed to the constructor of the :class:`Terminus`
            derived class.

        Raises
        ------
        :class:`~.TypeAttributeNotImplementedError`
            If the type attribute has not been defined in the derived class
            constructor. 

        :class:`~.TypeAttributeTypeError`
            If the :attr:`type` is not of type :obj:`type`

        :class:`~.SettingTypeError`
            If the type of the passed value is not :attr:`type`.

        """
        if not hasattr(self, "type"):
            raise TypeAttributeNotImplementedError(self.__class__)
        if not isinstance(self.type, type):
            raise TypeAttributeTypeError(self.__class__)
        if not isinstance(value, self.type):
            raise SettingTypeError(self.type, type(value))
        self.value = value


class List(Settings):
    """A base class for a list of settings classes.

    This class is used to store a number of settings objects. Instances must be
    of same type. 

    The implementation of the derived class constructor should be as follows::

        @Settings.assign
        def __init__(self, values: list):
            self.value = type
    
    Where `type` is the type of the objects that will be in the list.

    Attributes
    ----------
    type : :obj:`type`
        The attribute defined in the child class that determines the type of the
        setting stored in the class.

    value : :obj:`List`[:obj:`Any`]
        The list of value stored after all checks have been passed.

    """
    @property
    def get(self):
        return self.value

    def distribute(self, values: list):
        """Method called by the decorator :meth:`Settings.assign` that
        tries to assign the values passed to the constructor of the
        :class:`List` derived class.

        Parameters
        ----------
        values : :obj:`list`
            The values passed to the constructor of the :class:`List`
            derived class.

        Returns
        -------
        None

        Raises
        ------
        :class:`~.TypeAttributeNotImplementedError`
            If the type attribute has not been defined in the derived class
            constructor. 

        :class:`~.TypeAttributeTypeError`
            If the :attr:`type` is not of type :obj:`type`

        :class:`~.SettingTypeError`
            If `values` is not a :obj:`list`.
        
        :class:`~.SettingsListTypeError`
            If any of the items in `values` are not of the required type,
            as specified in the derived class.

        :class:`~.SettingsErrorMessage`
            If any exceptions are raised when instantiating any subsettings.

        """
        if not hasattr(self, "type"):
            raise TypeAttributeNotImplementedError(self.__class__)
        if not isinstance(self.type, type):
            raise TypeAttributeTypeError(self.__class__)

        if not isinstance(values, list):
            raise SettingTypeError(list, type(values))

        self.value = list()

        if self.type not in self.primitive:
            for idx, item in enumerate(values):
                try:
                    self.value.append(self.type(item))
                except SettingErrorMessage as e:
                    raise SettingErrorMessage(f"[{idx}]", e)
                except SettingTypeError as e:
                    raise SettingErrorMessage(f"[{idx}]", original_error=e)
                except SettingRangeTypeError as e:
                    raise SettingErrorMessage(f"[{idx}]", original_error=e)
                except SettingRangeKeyError as e:
                    raise SettingErrorMessage(f"[{idx}]", original_error=e)
                except ConsistencyError as e:
                    raise SettingErrorMessage(f"[{idx}]", original_error=e)
        else:
            for idx, item in enumerate(values):
                try:
                    if not isinstance(item, self.type):
                        raise SettingTypeError(self.type, type(item))
                    self.value.append(item)
                except SettingStringSelectionError as e:
                    raise SettingErrorMessage(f"[{idx}]", original_error=e)

    def __getitem__(self, key):
        """An overload of the list get item method.

        Returns the list or sub list or entry depending on `key`. If the items
        stored in the list are :class:`~.Terminus` derived, then it will return
        a list or value which are the `value` attribute of the terminus
        instances.

        """
        rv = self.value[key]
        if issubclass(self.type, Terminus):
            if isinstance(rv, Iterable):
                return [item.get for item in rv] 
            else:
                return rv.get
        else:
            return rv
    
    def __len__(self):
        """Mapping __len__ to attibute `value` __len__.
        
        Returns
        -------
        :obj:`int`
            The length of the stored list.

        """
        return len(self.value)


class Dict(Settings):
    """A base class for a dictionary of settings classes.

    This class is used to store a number of settings objects. Instances must be
    of same type. 

    The implementation of the derived class constructor should be as follows::

        @Settings.assign
        def __init__(self, values: dict):
            self.value = type
    
    Where `type` is the type of the objects that will be in the dictionary.

    Attributes
    ----------
    type : :obj:`type`
        The attribute defined in the child class that determines the type of the
        setting stored in the class.

    value : :obj:`Dict`[:obj:`str`, :obj:`Any`]
        The list of value stored after all checks have been passed.

    """

    @property
    def get(self):
        return self.value

    def distribute(self, values: dict):
        """Method called by the decorator :meth:`Settings.assign` that
        tries to assign the values passed to the constructor of the
        :class:`Dict` derived class.

        Parameters
        ----------
        values : :obj:`dict`
            The values passed to the constructor of the :class:`Dict`
            derived class.

        Returns
        -------
        None

        Raises
        ------
        :class:`~.TypeAttributeNotImplementedError`
            If the type attribute has not been defined in the derived class
            constructor. 

        :class:`~.TypeAttributeTypeError`
            If the :attr:`type` is not of type :obj:`type`

        :class:`~.SettingTypeError`
            If `values` is not a :obj:`list`.
        
        :class:`~.SettingsListTypeError`
            If any of the items in `values` are not of the required type,
            as specified in the derived class.

        :class:`~.SettingsErrorMessage`
            If any exceptions are raised when instantiating any subsettings.

        """
        if not hasattr(self, "type"):
            raise TypeAttributeNotImplementedError(self.__class__)
        if not isinstance(self.type, type):
            raise TypeAttributeTypeError(self.__class__)

        if not isinstance(values, dict):
            raise SettingTypeError(dict, type(values))

        self.value = dict()

        if self.type not in self.primitive:
            for key, value in values.items():
                try:
                    self.value[key] = self.type(value)
                except SettingErrorMessage as e:
                    raise SettingErrorMessage(key, e)
                except SettingRangeTypeError as e:
                    raise SettingErrorMessage(key, original_error=e)
                except SettingRangeKeyError as e:
                    raise SettingErrorMessage(key, original_error=e)
                except SettingStringSelectionError as e:
                    raise SettingErrorMessage(key, original_error=e)
                except ConsistencyError as e:
                    raise SettingErrorMessage(key, original_error=e)
        else:
            for key, value in values.items():
                try:
                    if not isinstance(value, self.type):
                        raise SettingTypeError(self.type, type(value))
                    self.value[key] = value
                except SettingTypeError as e:
                    raise SettingErrorMessage(key, original_error=e)
    
    def __getitem__(self, key):
        """An overload of the :obj:`dict` __getitem__ method.

        Returns the value for the given key. If the value is :class:`~.Terminus`
        derived, then it will return `value` attribute of the terminus instance.

        """
        rv = self.value[key]
        if Terminus in rv.__class__.__bases__:
            return rv.value
        else:
            return rv
    
    def __len__(self):
        """Mapping __len__ to attribute `value` __len__.
        
        Returns
        -------
        :obj:`int`
            The length of the stored list.

        """
        return len(self.value)
    
    def keys(self):
        return self.value.keys()
    
    def values(self):
        if issubclass(self.type, Terminus):
            return {k: v.value for k, v in self.value.items()}.values()
        else:
            return self.value.values()
    
    def items(self):
        if issubclass(self.type, Terminus):
            return {k: v.value for k, v in self.value.items()}.items()
        else:
            return self.value.items()


class StringSelection(Terminus):
    """The as class that stores a string from a predefined list.

    Attributes
    ----------
    type : :obj:`type`
        The attribute defined in the child class that determines the type of the
        setting stored in the class.

    value: :obj:`str`
        The stored string.
    
    options : :obj:`List`[:obj:`str`]
        The predefined list of string that are valid. This attribute should be
        defined in the child constructor.

    """
    
    @property
    def get(self):
        return self.value

    def distribute(self, value: str):
        """Method called by the decorator :meth:`Settings.assign` that
        tries to assign the values passed to the constructor of the
        :class:`Dict` derived class.

        Parameters
        ----------
        value: :obj:`str`
            The value to be stored.

        Raises
        ------
        :class:`~.OptionsAttributeNotImplementedError`
            If the :attr:`options` attribute has not been defined in the derived
            class constructor.

        :class:`~.OptionsAttributeTypeError`
            If the item contained in the :attr:`options` attribute are not
            strings.

        :class:`~.SettingTypeError`
            If the passed value `value` is not a :obj:`str`.

        :class:`~.SettingStringSelectionError`
            If the passed value `value` is not in :attr:`options`.

        """
        if not hasattr(self, "options"):
            raise OptionsAttributeNotImplementedError(self.__class__)
        if not isinstance(self.options, list):
            raise OptionsAttributeTypeError(self.__class__)
        if not isinstance(value, str):
            raise SettingTypeError(str, type(value))
        elif value not in self.options:
            raise SettingStringSelectionError(self.options)
        else:
            self.value = value


class Number(Terminus):
    """The special Terminus variant that is for numerical values.

    This class support range values in the form of arrays of min/max/num
    definitions.

    Attributes
    ----------
    value : :obj:`Union`[:obj:`float`, :obj:`int`, :obj:`List`]
        The value or list of values stored.
    
    """

    @property
    def is_range(self):
        """:obj:`bool` : True if the instance contains a range of values.
        
        """
        return self._range
    
    @property
    def match(self):
        """:obj:`Union`[None, :obj:`str`] : The match parameter.

        """
        return self._match

    def distribute(self, value):
        """Method called by the decorator :meth:`Terminus.assign` that
        tries to assign the values passed to the constructor of the
        :class:`Number` derived class.

        Parameters
        ----------
        value : :obj:`Union`[:obj:`float`, :obj:`int`, :obj:`dict`]
            
        Raises
        ------
        :class:`~.TypeAttributeNotImplementedError`
            If the type attribute has not been defined in the derived class
            constructor. 

        :class:`~.TypeAttributeTypeError`
            If the :attr:`type` is not of type :obj:`type`

        :class:`~.SettingTypeError`
            If `values` is not a number or valid range dictionary.
        
        """
        if not hasattr(self, "type"):
            raise TypeAttributeNotImplementedError(self.__class__)
        if not isinstance(self.type, type):
            raise TypeAttributeTypeError(self.__class__)

        if type(value) is self.type:
            self.__value(value)
        elif type(value) is dict and "array" in value:
            self.__array(value)
        elif type(value) is dict:
            self.__range(value)
        else:
            raise SettingTypeError(
                f"{self.type} || {{'array': [{self.type}]}} || "
                f"{{'min': {self.type}, 'max': {self.type}, 'num': {int}}}",
                type(value))
    
    def __value(self, value):
        """The method that assigns the attributes if a single value is passed.

        Parameters
        ----------
        value : :obj:`Union`[:obj:`float`, :obj:`int`]
            The value to be stored.

        """
        self.value = value
        self._range = False
        self._match = None
    
    def __array(self, value: dict):
        """The method that assigns the attributes if a array of value is passed.

        Parameters
        ----------
        value : :obj:`dict`[:obj:`str`, :obj:`list`]
            The dictionary with "array": listofvalues.

        Raises
        ------
        :class:`~.SettingTypeError`
            If the values in the array are not the same type as :attr:`type`

        """
        for item in value["array"]:
            if not isinstance(item, self.type):
                raise SettingTypeError(self.type, type(item))
        self.value = value["array"]
        self._range = True
        try:
            self._match = value["match"]
        except KeyError:
            self._match = None

    def __range(self, value: dict):
        """The method that assigns the attributes if a array of value is passed.

        Parameters
        ----------
        value : :obj:`dict`
            The dictionary with the range definition. Must be of the form::

                {
                    "max": value,
                    "min": value,
                    "num": int
                }
            
            Where value is of the defined type.

        Raises
        ------
        :class:`~.SettingRangeTypeError`
            If type(`value`["min"]) is not :attr`type`.

        :class:`~.SettingRangeKeyError`
            If `value`["min"] does not exist.

        :class:`~.SettingRangeTypeError`
            If type(`value`["max"]) is not :attr`type`.

        :class:`~.SettingRangeKeyError`
            If `value`["max"] does not exist.

        :class:`~.SettingRangeTypeError`
            If type(`value`["num"]) is not :obj:`int`. 

        :class:`~.SettingRangeKeyError`
            If `value`["num"] does not exist.

        """
        try:
            if not isinstance(value["min"], self.type):
                raise SettingRangeTypeError("min", self.type)
        except KeyError:
            raise SettingRangeKeyError("min")
        try:
            if not isinstance(value["max"], self.type):
                raise SettingRangeTypeError("max", self.type)
        except KeyError:
            raise SettingRangeKeyError("max")
        try:
            if not isinstance(value["num"], int):
                raise SettingRangeTypeError('num', int)
        except KeyError:
            raise SettingRangeKeyError("num")
        self.value = list(linspace(value['min'],
                                   value['max'],
                                   abs(value['num'])))
        self._range = True
        try:
            self._match = value["match"]
        except KeyError:
            self._match = None

    def lower_bound(self, value):
        """Checks if values are above or equal to a lower bound.
        
        Helper function to be called in derived class check implementation.

        Parameters
        ----------
        value : :obj:`Union`[:obj:`float`, :obj:`int`]
            The lower bound.

        Raises
        ------
        :obj:`ValueError`
            If any values are less than or equal to the provided bound.

        """
        if isinstance(self.value, self.type):
            if self.value < value:
                raise ValueError(f"must be >= {value}")
        elif isinstance(self.value, list):
            for item in self.value:
                if item < value:
                    raise ValueError(f"must be >= {value}")
        elif isinstance(self.value, dict):
            if self.value["min"] < value:
                raise ValueError(f"must be >= {value}")
            if self.value["max"] < value:
                raise ValueError(f"must be >= {value}")
    
    def upper_bound(self, value):
        """Checks if values are below or equal to an upper bound.
        
        Helper function to be called in derived class check implementation.

        Parameters
        ----------
        value : :obj:`Union`[:obj:`float`, :obj:`int`]
            The upper bound.

        Raises
        ------
        :obj:`ValueError`
            If any values are greater than or equal to the provided bound.

        """
        if isinstance(self.value, self.type):
            if self.value > value:
                raise ValueError(f"must be <= {value}")
        elif isinstance(self.value, list):
            for item in self.value:
                if item > value:
                    raise ValueError(f"must be <= {value}")
        elif isinstance(self.value, dict):
            if self.value["min"] > value:
                raise ValueError(f"must be <= {value}")
            if self.value["max"] > value:
                raise ValueError(f"must be <= {value}")
    
    def lower_bound_exclusive(self, value):
        """Checks if values are above an upper bound.
        
        Helper function to be called in derived class check implementation.

        Parameters
        ----------
        value : :obj:`Union`[:obj:`float`, :obj:`int`]
            The lower bound.

        Raises
        ------
        :obj:`ValueError`
            If any values are greater than the provided bound.

        """
        if isinstance(self.value, self.type):
            if self.value <= value:
                raise ValueError(f"must be > {value}")
        elif isinstance(self.value, list):
            for item in self.value:
                if item <= value:
                    raise ValueError(f"must be > {value}")
        elif isinstance(self.value, dict):
            if self.value["min"] <= value:
                raise ValueError(f"must be > {value}")
            if self.value["max"] <= value:
                raise ValueError(f"must be > {value}")
    
    def upper_bound_exclusive(self, value):
        """Checks if values are above an lower bound.
        
        Helper function to be called in derived class check implementation.

        Parameters
        ----------
        value : :obj:`Union`[:obj:`float`, :obj:`int`]
            The lower bound.

        Raises
        ------
        :obj:`ValueError`
            If any values are greater than the provided bound.

        """
        if isinstance(self.value, self.type):
            if self.value >= value:
                raise ValueError(f"must be < {value}")
        elif isinstance(self.value, list):
            for item in self.value:
                if item >= value:
                    raise ValueError(f"must be < {value}")
        elif isinstance(self.value, dict):
            if self.value["min"] >= value:
                raise ValueError(f"must be < {value}")
            if self.value["max"] >= value:
                raise ValueError(f"must be < {value}")
 

class ComputationalSpace:
    """Class that creates a tensor of settings objects from ranges.

    Given a settings object with :class:`~.Number` subsettings that are ranges,
    this class will construct a tensor where each element is a settings object
    with a particular value chosen for each range.

    Attributes
    ----------
    setting : :obj:`Type`[:class:`~.Settings`]
        The base settings object to be expanded.
    
    restrict : :obj:`dict`
        A dictionary of :obj:`str`: :obj:`str` pairs that are used to exclude
        subsetting branches from the exploration function for finding ranges.
        If when searching the settings object for ranges, a setting with the
        same name as a key in :attr:`restrict` is found, only subsettings with
        name equal to the corresponding key will be searched.
    
    """

    def __init__(self, setting: Type[Settings], restrict: StringDict):
        """The constructor for the :class:`ComputationalSpace` class.

        Parameters
        ----------
        setting : :obj:`Type`[:class:`~.Settings`]
            The settings object being expanded.

        restrict : :obj:`dict`[:obj:`str`, :obj:`str`]
            A dictionary of :obj:`str`: :obj:`str` pairs that are used to exclude
            subsetting branches from the exploration function for finding ranges.
            If when searching the settings object for ranges, a setting with the
            same name as a key in :attr:`restrict` is found, only subsettings with
            name equal to the corresponding key will be searched.

        """
        self.setting = setting
        self.restrict = restrict
        self.addresses = list()
        self.values = list()
        self.matched = dict()
        self.unmatched = list()
        self.space = list()
        self.explore(self.setting, list())
        self.build_space()

    def get_by_address(self, root: dict, address: StringList):
        return reduce(operator.getitem, address, root)

    def set_by_address(self, root: dict, address, value):
        self.get_by_address(root, address[:-1])[address[-1]] = value

    def explore(self, root, path, restrict = None):
        if issubclass(type(root), List):
            branch = enumerate(root.value)
        elif issubclass(type(root), Dict):
            branch = root.items()
        elif issubclass(type(root), Settings):
            branch = [(k, v) for k, v in root.__dict__.items() if "__" not in k]
        else:
            branch = list()
        if restrict:
            branch = [(k, v) for k, v in branch if k == restrict]
        for key, item in branch:
            new_path = deepcopy(path)
            if issubclass(type(item), Number):
                if item.is_range:
                    new_path.append(key)
                    if item.match:
                        if item.match in self.matched:
                            self.matched[item.match]["addresses"].append(new_path)
                            self.matched[item.match]["values"].append(item.get)
                        else:
                            self.matched[item.match] = {
                                "addresses": [new_path],
                                "values": [item.get] 
                            }
                    else:
                        self.unmatched.append(new_path)
                        self.addresses.append(new_path)
                        self.values.append(item.get)
            else:
                new_path.append(key)
                for k, v in self.restrict.items():
                    new_restrict = v if key == k else None
                    break
                self.explore(item, new_path, new_restrict)
   
    def build_space(self):
        for match, items in self.matched.items():
            self.addresses += items["addresses"]
            if len({len(i) for i in items["values"]}) != 1:
                warnings.warn(f"ranges with match id '{match}' have unequal "
                              f"length. Zipped to shortest.")
            self.values.append(list(zip(*items["values"])))
        for batch in itertools.product(*self.values):
            flat_batch = list()
            for item in batch:
                if isinstance(item, Iterable):
                    for subitem in item:
                        flat_batch.append(subitem)
                else:
                    flat_batch.append(item)
            rv = deepcopy(self.setting.__source__)
            for address, value in zip(self.addresses, flat_batch):
                self.set_by_address(rv, address, float(value))
            self.space.append(type(self.setting)(rv))

    def __getitem__(self, indices):
        if isinstance(indices, tuple):
            if len(indices) > len(self.values):
                raise IndexError(f"too many indices for array {self.shape}")
            for idx, item in enumerate(indices):
                if not isinstance(item, int):
                    raise IndexError("only integers are valid when accessing arrays")
                if not 0 <= item < self.shape[idx]:
                    raise IndexError(f"index {item} is out of bounds for axis "
                                     f"{idx} with size {self.shape[idx]}")
            index = 0
            for idx, item in enumerate(indices):
                index += item * int(prod([len(v) for v in self.values[idx + 1:]]))
        elif isinstance(indices, int):
            if len(self.values) > 1:
                raise IndexError(f"too many indices for array {self.shape}")
            if not 0 <= indices < self.shape[0]:
                raise IndexError(f"index {indices} is out of bounds for axis {0} "
                                 f"with size {self.shape[0]}")
            index = indices
        else:
            raise IndexError("only integers are valid when accessing arrays")

        return self.space[index]

    @property
    def shape(self):
        if len(self.values):
            return tuple((len(v) for v in self.values))
        else:
            return (1,)
        
    @property
    def zero(self) -> Union[Type[Settings], None]:
        if len(self.space) == 1:
            return self.space[0]
        else:
            return None
    
    @property
    def axes(self):
        for item, space in zip(self.addresses, self.values):
            print(self.build_path(item), space)

    def build_path(self, elements: StringList):
        rv = ""
        for item in elements:
            rv += f"{item} -> "
        return rv[:-4]
    
    def cout_summary(self):
        from textwrap import fill
        rv = ""
        rv += f"Computational space dimensions: " + f"{self.shape}\n"\
                .replace(",", " x")\
                .replace(")", "")\
                .replace("(", "")
        axis = 0
        for idx, item in enumerate(self.unmatched):
            rv += f"\naxis: {axis}:"
            rv += "\n"
            rv += f"\t{self.build_path(item)}"
            rv += "\n"
            rv += fill(f"\tvalues:  {self.values[idx]}\n",
                       width=70,
                       subsequent_indent="\t")
            axis += 1
        rv += "\n"
        for key, value in self.matched.items():
            rv += f"\naxis: {axis}:"
            rv += "\n"
            rv += f"\tmatch_id: {key}"
            for idx, item in enumerate(value["addresses"]):
                rv += "\n"
                rv += f"\t{self.build_path(item)}"
            rv += "\n"
            rv += fill(f"\tvalues: {list(zip(*value['values']))}",
                       width=70,
                       subsequent_indent="\t\t")
            axis += 1
        return rv
    
    def index(self, setting: Type[Settings]):
        linear_index = self.linear_index(setting) 
        return unravel_index(linear_index, self.shape)

    def linear_index(self, setting: Type[Settings]):
        return self.space.index(setting)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __len__(self):
        return len(self.space)


class SettingRangeKeyError(Error):
    """The exception raised when a :class:`~.Number` instance is missing an 
    entry in a range specification.

    """
    def __init__(self, key: str):
        """The constructor for the :class:`SettingRangeKeyError` class.

        """
        self.msg = f"No '{key}' parameter provided for range"


class SettingRangeTypeError(Error):
    """The exception raised when a :class:`~.Number` instance is passed an
    incorrect type in a range specification. 

    """
    def __init__(self, key: str, expected_type: T):
        """The constructor for the :class:`SettingRangeTypeError` class.

        """
        self.msg = f"The '{key}' parameter was not {expected_type}"


class SettingStringSelectionError(Error):
    """The exception raised when a :class:`~.StringSelection` instance is
    passed an invalid value.

    """
    def __init__(self, allowed: StringList):
        """The constructor for the :class:`SettingRangeTypeError` class.

        """
        self.msg = f"must be one of {allowed}"


class SettingNotFoundError(Error):
    """The exception raised when a setting cannot be found in the passed
    :obj:`dict`.

    """
    def __init__(self):
        """The constructor for the :class:`SettingNotFoundError` class.

        """
        self.msg = "Setting not found." 


class SettingCheckError(Error):
    """The exception raised when a :class:`Terminus` fails its value check.

    """
    def __init__(self, raised_exception, value: T):
        """The constructor for the :class:`SettingCheckError` class.

        Parameters
        ----------
        raised_exception 
            The exception raised by the test.
        value : T
            The value which failed the check.
        """
        self.msg = raised_exception
        

class SettingTypeError(Error):
    """The exception raised when the setting found in the passed :obj:`dict`
    is of the wrong type, when instantiating a :class:`~.Dict` object.

    """
    def __init__(self, expected: type, actual: type):
        """The constructor for the :class::`SettingDictTypeError` class.

        Parameters
        ----------
        expected : :obj:`type`
            The expected type.
        
        actual : :obj:`type`
            The received type.

        """
        self.msg = f"Expecting : {expected} | Received: {actual}"


class SettingErrorMessage(Error):
    """The exception raised when the setting found in the passed :obj:`dict`
    is of the wrong type.

    """
    def __init__(self,
                 current_name: str,
                 branch_error = None,
                 original_error: Union[SettingCheckError, SettingTypeError] = None):
        """The constructor for the :class::`SettingTypeError` class.

        Parameters
        ----------
        setting : :obj:`str`
            The name of the setting which has an incorrect type passed.

        setting_type : :obj:`type`
            The expected type of the setting.
        """
        if original_error:
            self.route = [current_name]
            self.original_error = original_error

        elif branch_error:
            self.route = [current_name] + branch_error.route
            self.original_error = branch_error.original_error
        else:
            raise ValueError("Must pass either new error or branch error as parameter.")

        super().__init__(self.build_message())
    
    def build_message(self):
        rv = str()
        join = " -> "
        for item in self.route:
            if "[" in item: 
                rv = rv[:-len(join)]
            rv += f"{item}" + join
        rv += f"{str(self.original_error.msg)}"
        return rv


class OptionsAttributeNotImplementedError(Error):
    """The exception raised when the `options` attribute has not been defined
    in a StringSelection derived class.

    """
    def __init__(self, derived_class):
        """The constructor for the :class::`OptionsAttributeNotImplementedError`
        class.

        Parameters
        ----------
        derived_class : :obj:`type`
            The name of the derived class that raised the exception.
        
        """
        self.msg = f"The {derived_class} class does not have a self.options"\
                   f"attribute defined in its constructor."
  

class OptionsAttributeTypeError(Error):
    """The exception raised when the value of the `options` attribute defined
    in a StringSelection derived class is not a :obj:`list`, or one of its
    elements is not a string.

    """
    def __init__(self, derived_class):
        """The constructor for the :class::`OptionsAttributeTypeError`
        class.

        Parameters
        ----------
        derived_class : :obj:`type`
            The name of the derived class that raised the exception.
        
        """
        self.msg = f"The {derived_class} class self.options attribute's value "\
                   f"is not of type {list}, or one of its elements is not of"\
                   f" type {str}"
        super().__init__(self.msg)
        

class TypeAttributeNotImplementedError(Error):
    """The exception raised when the `type` attribute has not been defined
    in a Terminus, Number, List, or Dict derived class.

    """
    def __init__(self, derived_class):
        """The constructor for the :class::`TypeAttributeNotImplementedError`
        class.

        Parameters
        ----------
        derived_class : :obj:`type`
            The name of the derived class that raised the exception.
        
        """
        self.msg = f"The {derived_class} class does not have a self.type "\
                   f"attribute defined in its constructor."
        super().__init__(self.msg)
 

class TypeAttributeTypeError(Error):
    """The exception raised when the value of the `type` attribute defined
    in a Terminus, Number, List, or Dict derived class is not a :obj:`type`.

    """
    def __init__(self, derived_class):
        """The constructor for the :class::`TypeAttributeTypeError`
        class.

        Parameters
        ----------
        derived_class : :obj:`type`
            The name of the derived class that raised the exception.
        
        """
        self.msg = f"The {derived_class} class self.type attribute's value is"\
                   f" not of type `type`."
        super().__init__(self.msg)


class ConsistencyError(Error):
    """The exception raised if a consistency check is failed.

    """

    def __init__(self, msg: str):
        """The constructor for the :class:`ConsistencyError` class

        Parameters
        ----------

        """
        self.msg = msg
        super().__init__(msg)



import unittest
import os
from os import listdir
from os.path import join
from os import remove
import shutil
from pathlib import Path

from ridt.config import ConfigFileParser

from ridt.container import Domain
from ridt.container.eddydiffusionrun import EddyDiffusionRun

from ridt.data import BatchDataStore


class ST24(unittest.TestCase):

    """System Test 24. Test the system
       is able to output the data store
       created during a batch run or
       single run mode to disk."""

    def setUp(self) -> None:

        this_dir = os.path.dirname(os.path.abspath(__file__))
        with ConfigFileParser(join(this_dir, "st23/config.json")) as cfp:
            self.c = cfp

        self.bds = BatchDataStore()
        self.domain = Domain(self.c)
        self.output_dir = join(this_dir, "st23/data")
        Path(self.output_dir).mkdir(parents=True, exist_ok=True)
        self.edr = EddyDiffusionRun(self.c, self.output_dir)

    def tearDown(self) -> None:
        shutil.rmtree(self.output_dir)

    def test_verify(self):
        output_types = ["concentration", "exposure"]
        geometries = ["points", "lines", "planes"]
        for output in output_types:
            for geometry in geometries:
                _path = join(self.output_dir, geometry, output, "data")
                for name, val in getattr(self.c.models.eddy_diffusion.monitor_locations, geometry).items():
                    self.assertIn(f"{name}.npy", listdir(_path))


if __name__ == "__main__":
    unittest.main()



import unittest
from os import listdir
from os.path import join
from os import remove
import shutil

from ridt.config import ConfigFileParser

from ridt.container import Domain
from ridt.container.eddydiffusionrun import EddyDiffusionRun

from ridt.data import BatchDataStore


class ST22(unittest.TestCase):

    """System Test 22. Test the system can
       produce animated contour plots
       where requested."""

    def setUp(self) -> None:
        pass

    def tearDown(self) -> None:
        pass

    def test_verify(self):
        pass


if __name__ == "__main__":
    unittest.main()



from os.path import join

from numpy import save

from ridt.config import RIDTConfig
from ridt.config import ConfigFileWriter

from .directoryagent import DirectoryAgent

from .datastore import DataStore


class DataStoreWriter:
    """A class that attempts to read a data store from disk.

    This will only work if the folder structure generated when writing the data
    store to disk is un modified.

    When creating a new instance of this class, an instance of
    :class:`~.DataStore` or :class:`~.BatchDataStore` will be returned instead
    of :class:`~.DataStoreReader`.

    Attributes
    ----------
    settings : :class:`~.RIDTConfig`
        The settings for the run output stored in the data store.
    
    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.

    dir_agent : :class:`~.DirectoryAgent`
        The path to the output directory for the run.

    """

    def __new__(cls, *args, **kwargs):
        instance = super(DataStoreWriter, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return instance

    def __init__(self,
                 setting: RIDTConfig,
                 data_store: DataStore,
                 dir_agent: DirectoryAgent,
                 quantity: str):
        """The :class:`~.DataStoreWriter` constructor.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run output stored in the data store.
        
        quantity: :obj:`str`
            The string id for the quantity stored in the data  store.

        dir_agent : :class:`~.DirectoryAgent`
            The path to the output directory for the run.

        data_store : :class:`~.DataStore`
            The data store to be written.

        """
        self.dir_agent = dir_agent
        self.setting = setting
        self.quantity = quantity
        self.write(data_store)
    
    @property
    def geometries(self):
        """:obj:`list` [:obj:`str`] : the list of geometries selected for
        evaluation in :attr:`settings`.

        """
        locations = self.setting.models.eddy_diffusion.monitor_locations
        return [g for g, e in locations.evaluate.items() if e]

    def write(self, data_store: DataStore) -> None:
        """Method that loop over the entries in the data store and writes them.

        The settings object is converted to JSON and written to disk.

        The :attr:`dir_agent` creates and provides paths to the relevant
        directories.

        Parameters
        ----------
        data_store : :class:`~.DataStore`
            The data store to be written.
        
        Returns
        -------
        None

        """
        ConfigFileWriter(self.dir_agent.outdir, "config.json", self.setting.__source__)
        for geometry in self.geometries:
            self.dir_agent.create_data_dir(geometry, self.quantity)
            for id in getattr(data_store, geometry):
                save(join(self.dir_agent.ddir, id), data_store.get(geometry, id))



import unittest

import os

from ridt.config import ConfigFileParser
from ridt.config.ridtconfig import ConsistencyError


class ST03(unittest.TestCase):

    """System Test 03. Test that the system is
    able to parse instantaneous/infinite duration/fixed
    duration sources from the JSON config file.."""

    def setUp(self) -> None:

        """setUp method which instantiates an
        :class:`~.RIDTConfig` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with ConfigFileParser(path) as cfp:
            self.c = cfp

    def test_verify(self):

        """Verifies that the :class:`~.RIDTConfig` class
        has the correct source attributes."""

        self.assertEqual(
            hasattr(self.c, "modes"), True)
        self.assertEqual(
            hasattr(self.c.modes, "instantaneous"), True)
        self.assertEqual(
            hasattr(self.c.modes, "infinite_duration"), True)
        self.assertEqual(
            hasattr(self.c.modes, "fixed_duration"), True)

    def test_instantaneous(self):

        """Verify that the :class:`~.InstantaneousSource`
        class has the correct attributes."""

        self.assertEqual(
            hasattr(self.c.modes.instantaneous, "sources"), True)
        for sources in self.c.modes.instantaneous.sources.values():
            self.assertEqual(hasattr(sources, "x"), True)
            self.assertEqual(hasattr(sources, "y"), True)
            self.assertEqual(hasattr(sources, "z"), True)
            self.assertEqual(hasattr(sources, "mass"), True)
            self.assertEqual(hasattr(sources, "time"), True)

    def test_infinite_duration(self):

        """Verify that the :class:`~.InfiniteDuration`
        class has the correct attributes."""

        self.assertEqual(
            hasattr(self.c.modes.infinite_duration, "sources"), True)
        for sources in self.c.modes.infinite_duration.sources.values():
            self.assertEqual(hasattr(sources, "x"), True)
            self.assertEqual(hasattr(sources, "y"), True)
            self.assertEqual(hasattr(sources, "z"), True)
            self.assertEqual(hasattr(sources, "rate"), True)
            self.assertEqual(hasattr(sources, "time"), True)

    def test_fixed_duration(self):

        """Verify that the :class:`~.FixedDuration`
        class has the correct attributes."""

        self.assertEqual(
            hasattr(self.c.modes.fixed_duration, "sources"), True)
        for sources in self.c.modes.fixed_duration.sources.values():
            self.assertEqual(hasattr(sources, "x"), True)
            self.assertEqual(hasattr(sources, "y"), True)
            self.assertEqual(hasattr(sources, "z"), True)
            self.assertEqual(hasattr(sources, "rate"), True)
            self.assertEqual(hasattr(sources, "start_time"), True)
            self.assertEqual(hasattr(sources, "end_time"), True)

    def test_source_times(self):

        """Checks to see if the :class:`.ConsistencyError` error
        triggers when the time of a source is greater than
        the total time of the simulation."""

        for key, value in self.c.modes.instantaneous.sources.items():
            value.time = self.c.total_time + 1
            self.assertRaises(ConsistencyError, self.c.consistency_check)

        for key, value in self.c.modes.infinite_duration.sources.items():
            value.time = self.c.total_time + 1
            self.assertRaises(ConsistencyError, self.c.consistency_check)

        for key, value in self.c.modes.fixed_duration.sources.items():
            value.start_time = self.c.total_time + 1
            self.assertRaises(ConsistencyError, self.c.consistency_check)

        for key, value in self.c.modes.fixed_duration.sources.items():
            value.start_time = value.end_time + 1
            self.assertRaises(ConsistencyError, self.c.consistency_check)


if __name__ == "__main__":
    unittest.main()



import unittest

import os

from ridt.base import ConsistencyError
from ridt.config import RIDTConfig
from ridt.config import ConfigFileParser
from ridt.config.configfileparser import ConfigFileParserValidationError


class ST01(unittest.TestCase):

    """System Test 01. The objective of this test
    is to prove that the system can parse Json files
    from disk and verify their integrity."""

    def setUp(self) -> None:

        """The set up method for system test 01.
        Configures a valid config file and a config
        file with an error in."""

        self.this_dir = os.path.dirname(os.path.abspath(__file__))
        with ConfigFileParser(os.path.join(self.this_dir, "../../default/config.json")) as cfp:
            self.c = cfp

    def test_parse(self):

        """Checks to see that the consistency check can raise
        an error in a certain circumstance."""

        self.c.consistency_check()
        with self.assertRaises(ConsistencyError):
            ConfigFileParser(os.path.join(self.this_dir, "st01/error_config.json"))

    def test_verify(self):

        """Verifies that the config file parser outputs a
        :class:`~.RIDTConfig` class."""

        self.assertEqual(
            type(self.c), RIDTConfig
        )


if __name__ == "__main__":
    unittest.main()



import unittest
import json
import os

import numpy as np
from matplotlib.lines import Line2D

from ridt.config import RIDTConfig
from ridt.plot import LinePlot


class TestLinePlot(unittest.TestCase):

    "Unit tests for the :class:`~.LinePlot` class."""

    def setUp(self) -> None:

        """setUp method which instantiates the :class:`~.RIDTConfig` class,
        the :class:`~.LinePlot` class, and creates some initial
        variables."""

        self.this_dir = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(self.this_dir, "test_resources/test_config.json")) as f:
            loaded_json = json.load(f)

        self.config = RIDTConfig(loaded_json)

        self.lp = LinePlot(self.config, os.path.join(self.this_dir, "test_resources", "concentration"))

        self.time_array = np.linspace(0, 10, 10)
        self.conc = np.linspace(0, 10, 10)
        self.lines = self.config.models.eddy_diffusion.monitor_locations.lines

    def tearDown(self) -> None:

        """tearDown method which removes and changes made in the
        tests."""

        resource_list = os.listdir(os.path.join(self.this_dir, "test_resources"))
        for file in resource_list:
            if file.endswith(".pdf"):
                os.remove(os.path.join(self.this_dir, f"test_resources/{file}"))


if __name__ == "__main__":
    unittest.main()



import unittest

import os

from ridt.config import ConfigFileParser
from ridt.config.ridtconfig import ConsistencyError


class ST07(unittest.TestCase):

    """System Test 07. Test that the system can
       parse specified concentration and exposure
       threshold levels from the JSON file."""

    def setUp(self) -> None:

        """setUp method that instantiates the
        :class:`~.RIDTConfig` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with ConfigFileParser(path) as cfp:
            self.c = cfp

    def test_verify(self):

        """Verify that :class:`~.Thresholds` class contains
        the threshold attributes."""

        self.assertEqual(
            hasattr(self.c, "thresholds"), True
        )
        self.assertEqual(
            hasattr(self.c.thresholds, "concentration"), True
        )
        self.assertEqual(
            hasattr(self.c.thresholds, "exposure"), True
        )

    def test_thresholds(self):

        """Checks to see if the :class:`~.ConsistencyError` error
        triggers when the number of thresholds exceeds 5."""

        for i in range(6):
            self.c.thresholds.concentration.append(1)
        self.assertRaises(ConsistencyError, self.c.consistency_check)

        for i in range(6):
            self.c.thresholds.exposure.append(1)
        self.assertRaises(ConsistencyError, self.c.consistency_check)


if __name__ == "__main__":
    unittest.main()



import unittest
import json
import os

from ridt.config import RIDTConfig
from ridt.base import ConsistencyError


class TestConsistencyChecker(unittest.TestCase):


    def setUp(self) -> None:

        """setUp method which instantiates the
        :class:`~.RIDTConfig` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with open(path) as f:
            loaded_json = json.load(f)

        self.config = RIDTConfig(loaded_json)


if __name__ == "__main__":
    unittest.main()



import unittest
import json
import os

import numpy as np

from ridt.config import RIDTConfig
from ridt.equation import WellMixed


class TestWellMixed(unittest.TestCase):

    def setUp(self) -> None:

        this_dir = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(this_dir, "test_resources/test_config.json")) as f:
            loaded_json = json.load(f)

        self.config = RIDTConfig(loaded_json)

        self.wm = WellMixed(self.config)

    def test_concentration(self):

        """Makes sure that the concentration function outputs
        the correct type."""

        for time in range(10):
            exp = self.wm.concentration(time)

            self.assertEqual(type(exp), np.float64)

    def test_outputs(self):

        """Ensures that the concentration output type
        is ndarray."""

        time_array = np.linspace(0, 10, 10)
        inst_conc = self.wm(time_array)
        self.assertEqual(type(inst_conc), np.ndarray)



import unittest
import os
import shutil
from pathlib import Path

from ridt.base import ComputationalSpace

from ridt.config import ConfigFileParser

from ridt.data.batchdatastore import BatchDataStoreIDError
from ridt.data import DataStore
from ridt.data import BatchDataStore
from ridt.data import BatchDataStoreWriter

from ridt.analysis import DataStoreAnalyser
from ridt.analysis import BatchDataStoreAnalyser
from ridt.analysis.batchresultswriter import BatchResultsWriter


class ST14(unittest.TestCase):

    """System Test 14. Test the system can
       provide summary output for a given
       configuration file."""

    def setUp(self) -> None:

        """setUp method that instantiates the
           :class:`~.RIDTConfig` class and the
           :class:`~.ComputationalSpace` class."""

        self.this_dir = os.path.dirname(os.path.abspath(__file__))
        self.out_dir = os.path.join(self.this_dir, "st14/run")
        Path(self.out_dir).mkdir(parents=True, exist_ok=True)

        with ConfigFileParser(os.path.join(self.this_dir, "st14/config.json")) as cfp:
            self.c = cfp

        restrict = {"models": "well_mixed"}
        self.space = ComputationalSpace(self.c, restrict)

    def tearDown(self) -> None:
        shutil.rmtree(self.out_dir)

    def test_verify(self):

        """Verifies that the config.json file is created
        in the correct directory, then removes the created
        files."""

        da = DataStore()
        dsa = DataStoreAnalyser(self.c, da, "concentration")

        bds = BatchDataStore()
        bds.add_run(self.c)

        bdsw = BatchResultsWriter(
            self.c, self.space, {self.c: dsa}, self.out_dir, "concentration"
        )

        try:
            bdsw.write()
        except BatchDataStoreIDError as e:
            pass
        self.assertIn(
            "batch_run_summary.txt", os.listdir(f"{self.out_dir}")
        )

        with open(os.path.join(self.this_dir, "st14/run/batch_run_summary.txt")) as f:
            txt = f.read()

        self.assertIn(
            "Computational space dimensions: 1 x",
            txt
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import os

from ridt.base import ComputationalSpace
from ridt.config import ConfigFileParser
from ridt.config import RIDTConfig


class ST11(unittest.TestCase):

    """System Test 11. Test the system can convert
       config files with ranges of values into a set of
       config object, each which can be evaluated separately."""

    def setUp(self) -> None:

        """setUp method that instantiates the
           :class:`~.RIDTConfig` class and the
           :class:`~.ComputationalSpace` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "st11/config.json")
        with ConfigFileParser(path) as cfp:
            self.c = cfp

            restrict = {"models": "well_mixed"}
            self.space = ComputationalSpace(self.c, restrict)

    def test_verify(self):

        settings_list = list()
        for setting in self.space.space:
            if isinstance(setting, RIDTConfig):
                settings_list.append(setting)
        self.assertEqual(len(settings_list) == 9, True)


if __name__ == "__main__":
    unittest.main()



import unittest
import os
from os import listdir
from os.path import join
from os import remove
from os import system
import shutil
import json
from pathlib import Path

from ridt.config import ConfigFileParser

from ridt.config.ridtconfig import FixedDurationSource
from ridt.config.ridtconfig import InstantaneousSource
from ridt.config.ridtconfig import InfiniteDurationSource


class ST27(unittest.TestCase):
    """System Test 27. Test the system can
       provide access to all functionality
       via a command line interface."""

    def setUp(self) -> None:

        self.this_dir = os.path.dirname(os.path.abspath(__file__))
        self.config_path = join(self.this_dir, "st27/config.json")
        self.csv_path = join(self.this_dir, "st27/csv_to_config.csv")

        self.output_dir = join(self.this_dir, "st27/data")
        Path(self.output_dir).mkdir(parents=True, exist_ok=True)
        self.virtualenv_path = ".venv"

    def tearDown(self) -> None:
        shutil.rmtree(self.output_dir)

        with open(self.config_path) as f:
            config = json.load(f)

        csv_config = config
        with open(self.config_path, "w") as w:
            json.dump(csv_config, w)
        w.close()

    def test_run(self):
        try:
            system(f"{self.virtualenv_path}\\Scripts\\activate")
            system(f"ridt run {self.config_path} {self.output_dir}")

            output = False
            if listdir(self.output_dir):
                output = True

            self.assertEqual(True, output)

        except Exception:
            print("ridt failure")
            self.assertEqual(True, False)

    def test_csv_to_config(self):
        try:
            modes = ["instantaneous", "infinite_duration", "fixed_duration"]

            ins = {"x": 1.0, "y": 1.0, "z": 1.0, "mass": 1.0, "time": 1.0}
            inf = {"x": 1.0, "y": 1.0, "z": 1.0, "rate": 1.0, "time": 1.0}
            fix = {"x": 1.0, "y": 1.0, "z": 1.0, "rate": 1.0, "start_time": 1.0, "end_time": 2.0}

            system(f"{self.virtualenv_path}\\Scripts\\activate")
            system(f"ridt csv-to-config {self.config_path} {self.csv_path}")

            config = ConfigFileParser(os.path.join(self.this_dir, "st27/new_config.json"))
            for mode in modes:
                m = getattr(config.modes, mode)
                for source in m.sources.values():
                    if isinstance(source, InstantaneousSource):
                        self.assertEqual(source.__source__, ins)
                    if isinstance(source, InfiniteDurationSource):
                        self.assertEqual(source.__source__, inf)
                    if isinstance(source, FixedDurationSource):
                        self.assertEqual(source.__source__, fix)
        except Exception:
            print("csv_to_config failure")
            self.assertEqual(True, False)


if __name__ == "__main__":
    unittest.main()



from os import mkdir

from os.path import join

from numpy import unravel_index

from ridt.base import RIDTOSError


class DirectoryAgent:
    """This class provides directory creation functionality.

    This class provides various methods for creating nested directory structures
    required to store the raw binary grids, csv files, analysis, and plots.

    Attributes
    ----------
    rootdir : :obj:`str`
        The main directory where everything is stored.

    outdir : :obj:`str`
        The the current root dir for a given computational space index.

    gdir: :obj:`str`
        The the current geometry dir relative to the :attr:`rootdir`.
    
    qdir : :obj:`str`
        The current quantity dir relative to the :attr:`gdir`.
    
    ddir : :obj:`str`
        The current data dir relative to the :attr:`qdir`.

    pdir : :obj:`str`
        The current plot dir relative to the :attr:`qdir`.
    
    adir : :obj:`str`
        The current analysis dir relative to the :attr:`qdir`.
    
    shape : :obj:`Tuple`[:obj:`int`]
        The shape of the computational space.

    """
    
    def __init__(self, outdir: str, shape: tuple):
        """The :class:`~.DirectoryAgent` constructor.

        Parameters
        ----------


        """
        self.shape = shape
        self.rootdir = outdir
        self.outdir = outdir
        self.gdir = None
        self.qdir = None
        self.ddir = None
        self.pdir = None
        self.adir = None
    
    def create_root_dir(self, run_idx: tuple) -> None:
        """Create the subdirectory for a batch run element.

        Parameters
        ----------
        run_idx : :obj:`Tuple`[:obj:`int`]
            The index in computational space of the run.
        
        Returns
        -------
        None

        """
        idx = unravel_index(run_idx, self.shape)
        idx = str(idx).replace("(","[").replace(")", "]")
        self.outdir = join(self.rootdir, idx)
        self.mkdir(self.outdir)
   
    def create_geometry_dir(self, geometry: str):
        """Create the geometry subdirectory.

        Parameters
        ----------
        geometry : :obj:`str`
            The name of the geometry class.

        Returns
        -------
        :obj:`str`
            The path to the new directory.

        """
        self.gdir = join(self.outdir, geometry)
        self.mkdir(self.gdir)
        return self.gdir

    def create_quantity_dir(self, geometry: str, quantity: str):
        """Create the quantity subdirectory.

        Parameters
        ----------
        geometry : :obj:`str`
            The name of the geometry class.
        
        quantity : :obj:`str`
            The string id of the quantity.  

        Returns
        -------
        :obj:`str`
            The path to the new directory.

        """
        self.create_geometry_dir(geometry)
        self.qdir = join(self.gdir, quantity)
        self.mkdir(self.qdir)
        return self.qdir
    
    def create_data_dir(self, geometry: str, quantity: str):
        """Create the data subdirectory.

        Parameters
        ----------
        geometry : :obj:`str`
            The name of the geometry class.
        
        quantity : :obj:`str`
            The string id of the quantity.  

        Returns
        -------
        :obj:`str`
            The path to the new directory.

        """
        self.create_quantity_dir(geometry, quantity)
        self.ddir = join(self.qdir, "data")
        self.mkdir(self.ddir)
        return self.ddir
    
    def create_plot_dir(self, geometry: str, quantity: str):
        """Create the plots subdirectory.

        Parameters
        ----------
        geometry : :obj:`str`
            The name of the geometry class.
        
        quantity : :obj:`str`
            The string id of the quantity.  

        Returns
        -------
        :obj:`str`
            The path to the new directory.

        """
        self.create_quantity_dir(geometry, quantity)
        self.pdir = join(self.qdir, "plots")
        self.mkdir(self.pdir)
        return self.pdir
    
    def create_analysis_dir(self, geometry: str, quantity: str):
        """Create the analysis subdirectory.

        Parameters
        ----------
        geometry : :obj:`str`
            The name of the geometry class.
        
        quantity : :obj:`str`
            The string id of the quantity.  

        Returns
        -------
        :obj:`str`
            The path to the new directory.

        """
        self.create_quantity_dir(geometry, quantity)
        self.adir = join(self.qdir, "analysis")
        self.mkdir(self.adir)
        return self.adir

    def mkdir(self, path: str) -> None:
        """Create a directory at the provided path.

        Parameters
        ----------
        path : :obj:`str`
            The path to the new directory.
        
        Returns
        -------
        None

        Raises
        ------
        :class:`~.RIDTOSError`
            If cannot create the new directory.

        """
        try:
            mkdir(path)
        except FileExistsError:
            pass
        except OSError as e:
            raise RIDTOSError(e)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



from .batchdatastore import BatchDataStore

from .batchdatastorewriter import BatchDataStoreWriter

from .datastore import DataStore

from .datastorewriter import DataStoreWriter

from .directoryagent import DirectoryAgent

from .datastoreplotter import DataStorePlotter

from .batchdatastoreplotter import BatchDataStorePlotter

from .datastorereader import DataStoreReader

from .uncertaintymask import UncertaintyMask


import json

from os.path import join

from ridt.base import Error


class ConfigFileWriter:
    """The class which handles the configuration file writing.

    This class handles the writing of configuration files from disk, as well as
    the instantiation of the :class:`~.RIDTConfig` instances using the written
    data.

    Attributes
    ----------
    path : :obj:`str`
        The path to the file in which to write the configuration JSON.
    
    data : :obj:`dict`
        The configuration stored as a :obj:`dict`.

    """
    def __new__(cls, *args, **kwargs):
        instance = super(ConfigFileWriter, cls).__new__(cls)
        instance.__init__(*args, **kwargs)

    def __init__(self, output_dir: str, file_name: str, data: dict):
        self.path = join(output_dir, file_name)
        self.data = data
        self.__write_file()
       
    def __write_file(self):
        """The method which writes the configuration file.

        Returns
        -------
        None

        Raises
        ------
        :class:`~.ConfigFileWriterJSONError`
            If the structure of the written JSON is invalid.
        
        :class:`~.ConfigFileWriterOSError`
            If an :obj:`OSError` is raised when writing the file.

        """

        try:
            with open(self.path, 'w') as f:
                f.write(json.dumps(self.data, indent=4))
        except OSError as e:
            raise ConfigFileWriterOSError(self.path, e)


class ConfigFileWriterOSError(Error):
    """The exception raised when a :class:`~.ConfigFileWriter` instance
    catches an error when writing a config file path, or instantiating the
    :class:`~.Settings` derived class.

    """
    def __init__(self, path: str, error: OSError):
        """The constructor for the :class:`ConfigFileWriterOSError` class.

        """
        msg = f"While loading {path}, I encountered the following error:"\
              f"\n {error}"
        super().__init__(msg)



import unittest

import os

from ridt.config import ConfigFileParser


class ST08(unittest.TestCase):

    """System Test 08. Test that the system can
       parse the options relating to the number of
       image sources and allowable error for the
       Eddy diffusion model from the JSON file."""

    def setUp(self) -> None:

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with ConfigFileParser(path) as cfp:
            self.c = cfp

    def test_verify(self):

        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion, "images"), True
        )
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.images, "quantity"), True
        )


if __name__ == "__main__":
    unittest.main()



import sys

import shutil

from os.path import isdir
from os.path import join
from os.path import dirname
from os import getcwd
from os import listdir


import click

from ridt.base import ConsistencyError
from ridt.config.configfileparser import ConfigFileParser
from ridt.config.configfileparser import ConfigFileParserJSONError
from ridt.config.configfileparser import ConfigFileParserOSError
from ridt.config.configfileparser import ConfigFileParserValidationError
from ridt.config.csvtoconfigfile import CSVToConfigFile
from ridt.container.wellmixedrun import WellMixedRun
from ridt.container.eddydiffusionrun import EddyDiffusionRun
from ridt.base import RIDTOSError

@click.group()
def ridt():
    """The rapid indoor diffusion tool (ridt)."""
    pass

@ridt.command()
def init():
    """Copy a default config file to current working directory.

    """
    try:
        source_path = join(dirname(__file__), "..", "default")
        for item in [f for f in listdir(source_path) if "__init__.py" not in f]:
            s = join(source_path, item)
            d = join(getcwd(), item)
            if isdir(s):
                shutil.copytree(s, d)
            else:
                shutil.copy2(s, d)
    except OSError as e:
        print(f"Could not clone default config file. Error: {e}")


@ridt.command()
@click.argument('config_file', type=click.Path(exists=True))
@click.argument('output_dir', type=click.Path(exists=True))
def run(config_file, output_dir):
    """Run diffusion model."""

    if not isdir(output_dir):
        sys.exit(f"{output_dir} is not a directory.\n\nAborted.")

    try:
        s = ConfigFileParser(config_file)
    except (ConfigFileParserJSONError,
            ConfigFileParserOSError,
            ConfigFileParserValidationError) as e:
        sys.exit(e)
    except ConsistencyError as e:
        sys.exit(e)

    try:
        if s.well_mixed:
            WellMixedRun(s, output_dir)
        if s.eddy_diffusion:
            EddyDiffusionRun(s, output_dir)
    except RIDTOSError as e:
        sys.exit(f"\n{e}\n\nAborted.")
    
    print("\nComplete.")


@ridt.command()
@click.argument('config_file', type=click.Path(exists=True))
@click.argument('csv_file', type=click.Path(exists=True))
@click.option('-o', '--output_file', type=click.Path())
@click.option('--force/--no-force', default=False)
def csv_to_config(config_file, csv_file, output_file, force):
    """Merge CSV file to config JSON file. """
    if config_file == output_file:
        if not force:
            sys.exit("To overwrite the original config file, use the --force"
                     " flag.")
    with CSVToConfigFile() as ctc:
        ctc(config_file, csv_file, output_file)

    pass



import unittest
import os
from os import listdir
from os.path import join
from os import remove
from pathlib import Path

import shutil

from ridt.config import ConfigFileParser

from ridt.container import Domain
from ridt.container.eddydiffusionrun import EddyDiffusionRun

from ridt.data import BatchDataStore
from ridt.data import BatchDataStoreWriter


class ST18(unittest.TestCase):

    """System Test 18. test the system can
       write all data output by the models to
       a given directory."""

    def setUp(self) -> None:

        this_dir = os.path.dirname(os.path.abspath(__file__))
        with ConfigFileParser(os.path.join(this_dir, "st16/explicit.json")) as cfp:
            self.c = cfp

        self.bds = BatchDataStore()
        self.domain = Domain(self.c)
        self.output_dir = os.path.join(this_dir, "st18")
        Path(self.output_dir).mkdir(parents=True, exist_ok=True)
        self.edr = EddyDiffusionRun(self.c, self.output_dir)
        self.list_dir = listdir(self.output_dir)

    def tearDown(self) -> None:
        shutil.rmtree(self.output_dir)

    def test_write(self):

        geometries = ["points", "lines", "planes"]
        output_types = ["concentration", "exposure"]

        for geometry in geometries:
            for output in output_types:
                _dir = f"{geometry}/{output}/data"
                _path = join(self.output_dir, _dir)
                for val in self.edr.data_store.store.values():
                    for name, monitor in getattr(val, geometry).items():
                        self.assertIn(f"{name}.npy", listdir(_path))
                        self.assertIn(f"{name}.csv", listdir(_path))


if __name__ == "__main__":
    unittest.main()



import unittest

import os

import json

from ridt.config import ConfigFileParser
from ridt.config.ridtconfig import ConsistencyError

from ridt.config.configfileparser import ConfigFileParserValidationError


class ST11(unittest.TestCase):

    """System Test 11. Test the system can convert
       config files with ranges of values into a set of
       config object, each which can be evaluated separately."""

    def setUp(self) -> None:

        """setUp method that instantiates the
           :class:`~.RIDTConfig` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        self.config_path = os.path.join(this_dir, "st06/config.json")
        with open(self.config_path) as f:
            self.default = json.load(f)

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with ConfigFileParser(path) as cfp:
            self.c = cfp

    def tearDown(self) -> None:

        with open(self.config_path, "w") as w:
            json.dump(self.default, w)

    def test_verify(self):

        """Verify that the :class:`~.ContourPlots` class
        has the correct scale and range attributes."""

        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.planes_plots, "range"), True
        )
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.planes_plots, "scale"), True
        )

    def test_contours(self):

        with self.assertRaises(ConfigFileParserValidationError):
            with open(self.config_path) as f:
                config = json.load(f)
            config["models"]["eddy_diffusion"]["planes_plots"]["contours"]["min"] = \
                config["models"]["eddy_diffusion"]["planes_plots"]["contours"]["max"] + 1
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)

    def test_range(self):

        with self.assertRaises(ConfigFileParserValidationError):
            with open(self.config_path) as f:
                config = json.load(f)
            config["models"]["eddy_diffusion"]["planes_plots"]["range"] = "test"
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)

    def test_scale(self):

        with self.assertRaises(ConfigFileParserValidationError):
            with open(self.config_path) as f:
                config = json.load(f)
            config["models"]["eddy_diffusion"]["planes_plots"]["scale"] = "test"
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)


if __name__ == "__main__":
    unittest.main()



import warnings

from os.path import join

from numpy import ndarray
from numpy import log10
from numpy import floor
from numpy import ceil 
from numpy import power
from numpy import linspace
from numpy import transpose

from ridt.config import RIDTConfig
from ridt.config import Units 

from ridt.container import Domain

from matplotlib import cm, ticker
from matplotlib import colors

from matplotlib.ticker import LogFormatter
from matplotlib.ticker import FormatStrFormatter
import matplotlib.pyplot as plt
from matplotlib.pyplot import figure


class ContourPlot:
    """The class that generates contour plots.

    Attributes
    ----------
    settings : :class:`~.RIDTConfig`
        The settings for the run in question.

    domain : :class:`~.Domain`
        The instance of :class:`~.Domain` corresponding to :attr:`setting`.

    units : :class:`~.Units`
        The instance of :class:`~.Units` corresponding to :attr:`setting`.
    
    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.
    
    config : :class:`~.ContourPlots`
        The contour plots settings object.

    outdir: :obj:`str`
        The path to the output directory for the run.

    """
    def __init__(self, settings: RIDTConfig, output_dir: str, quantity: str):
        """The :class:`~.ContourPlot` constructor.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.

        outdir : :obj:`str`
            The path to the output directory for the run.

        quantity : :obj:`str`
            The string id for the quantity stored in the data  store.

        """
        self.settings = settings
        self.output_dir = output_dir
        self.units = Units(settings)
        self.domain = Domain(self.settings)
        self.quantity = quantity
        self.config = self.settings.models.eddy_diffusion.planes_plots

    def __call__(self, id: str, data: ndarray, max_val: float, t_index: int) -> None:
        """The call method used to plot some data.

        Parameters
        ----------
        id : :obj:`str`
            The id string of the plane being plotted.

        data : :class:`~numpy.ndarray`
            The array of values to be plot.

        max_val : :obj:`float`
            The maximum value over the whole time domain.

        t_index : :obj:`int`
            The time index of the current data array.

        """
        self.id = id
        self.t_index = t_index
        self.max_val = max_val
        self.xaxis, self.yaxis = self.get_axes()
        self.figsize = self.get_figsize()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.plot(data)
        self.save_fig()
        plt.close()

    def plot(self, data: ndarray):
        """Calls various plotting functions.

        Parameters
        ----------
        data : :class:`~numpy.ndarray`
            The array of values to be plot.

        """

        figure(num=None, figsize=self.figsize, dpi=120, facecolor='w', edgecolor='k')

        plt.title(self.title())
        plt.xlabel(self.xlabel())
        plt.ylabel(self.ylabel())

        if self.config.scale == "logarithmic":
            levels = self.get_log_scale()
            kwargs = {"extend": "both", "norm": colors.LogNorm()}
        else:
            levels = self.get_linear_scale()
            kwargs = {}

        plt.contourf(
            self.get_xdomain(),
            self.get_ydomain(),
            transpose(data), 
            levels,
            cmap=cm.RdBu_r,
            **kwargs)

        if self.config.scale == "logarithmic":
            formatter = LogFormatter(10, labelOnlyBase=False, minor_thresholds=(1, 5))
            cb = plt.colorbar(ticks=self.get_log_scale(), format=formatter, )
            cb.ax.yaxis.set_major_formatter((FormatStrFormatter('%.2e')))
        else:
            plt.colorbar()

        plt.tight_layout()

    def save_fig(self):
        """Creates the formatted string file name and saves plot to disk.

        """
        plt.savefig(join(self.output_dir, f"{self.id}-{self.domain.time[self.t_index]:.2f}s.png"))

    def title(self):
        """Creates formatted title string for the current plot.

        Returns
        -------
        :obj:`str`
            The title string.

        """
        u = getattr(self.units, f'{self.quantity}')
        return f"{self.quantity} ({u}) - {self.id} - {self.domain.time[self.t_index]:.2f}s"
    
    def xlabel(self):
        """Creates the formatted x-axis label string.

        Returns
        -------
        :obj:`str`
            The x-axis label string.

        """
        return f"{self.xaxis} ({self.units.space})"

    def ylabel(self):
        """Creates the formatted y-axis label string.

        Returns
        -------
        :obj:`str`
            The y-axis label string.

        """
        return f"{self.yaxis} ({self.units.space})"

    def get_xdomain(self):
        """Extract the x-axis values from the :attr:`domain`

        Returns
        -------
        :obj:`Iterable`[:obj:`float`]
            The x-axis values.

        """
        return getattr(self.domain, self.xaxis)

    def get_ydomain(self):
        """Extract the x-axis values from the :attr:`domain`

        Returns
        -------
        :obj:`Iterable`[:obj:`float`]
            The x-axis values.

        """
        return getattr(self.domain, self.yaxis)
    
    def get_figsize(self):
        """Computes the aspect ratio for the plot.

        This is done relative the the physical bounds of the system.

        Returns
        -------
        :obj:`~.Tuple`[:obj:`float`]
            The width/height aspect ratio values.

        """
        dim = self.settings.dimensions
        aspect_ratio = getattr(dim, self.xaxis) / getattr(dim, self.yaxis)
        # 8 x 6 is the default values for the fig size in matplotlib.
        return (8 * aspect_ratio, 6)

    def get_axes(self):
        """Parse the x and y plot axes from the plane settings object.

        Returns
        -------
        :obj:`~.Tuple`[:obj:`str`]
            The two axes for the plot (x, y), (y, z), (x, z)

        """
        planes = self.settings.models.eddy_diffusion.monitor_locations.planes
        string = planes[self.id].axis
        return string[0], string[1]
    
    def get_log_scale(self):
        """Generate log scale values for the contours.

        This is either performed automatically or relative to defined values,
        depending on the user settings.

        Returns
        -------
        :obj:`Iterable`[:obj:`float`]
            The log scale values.
        
        """
        num_contour = self.config.number_of_contours    
        if self.config.range == "auto":
            min_contour = floor(log10(1e-20)-1)
            max_contour = ceil(log10(self.max_val))
        else:
            min_contour = floor(log10(self.config.contours.min))
            max_contour = ceil(log10(self.config.contours.max))
        levels = linspace(min_contour, max_contour, num_contour)
        return power(10, levels)

    def get_linear_scale(self):
        """Generate linear values for the contours.

        This is either performed automatically or relative to defined values,
        depending on the user settings.

        Returns
        -------
        :obj:`Iterable`[:obj:`float`]
            The linear scale values.
        
        """
        if self.config.range == "auto":
            min_contour = 0.0
            max_contour = self.max_val
        else:
            min_contour = self.config.contours.min
            max_contour = self.config.contours.max
        return linspace(min_contour, max_contour, self.config.number_of_contours)



import unittest
import os
from os import listdir
from os.path import join
from os import remove
import shutil
from numpy import squeeze
from pathlib import Path

from ridt.config import ConfigFileParser

from ridt.equation import EddyDiffusion

from ridt.container import Domain
from ridt.container.eddydiffusionrun import EddyDiffusionRun

from ridt.data import BatchDataStore


class ST17(unittest.TestCase):

    """System Test 17. test the system can use
       the information in the parsed JSON configuration
       file to evaluate the specified Eddy Diffusion model
       for all times, monitor locations, full domain, and
       points in parameter space (as defined by the ranges
       in the configuration file)."""

    def setUp(self) -> None:

        this_dir = os.path.dirname(os.path.abspath(__file__))
        self.output_dir = os.path.join(this_dir, "st17")
        Path(self.output_dir).mkdir(parents=True, exist_ok=True)
        with ConfigFileParser(os.path.join(this_dir, "st16/explicit.json")) as cfp:
            self.c = cfp

        self.bds = BatchDataStore()
        self.domain = Domain(self.c)
        self.edr = EddyDiffusionRun(self.c, self.output_dir)

    def tearDown(self) -> None:
        shutil.rmtree(self.output_dir)

    def test_verify(self):

        for val in self.edr.data_store.store.values():
            for point in val.points.values():
                self.assertEqual(len(point), self.c.time_samples)
            for line_name, line in val.lines.items():
                monitor = self.c.models.eddy_diffusion.monitor_locations.lines[line_name]
                axis = getattr(monitor, "parallel_axis")
                shape = (self.c.time_samples, getattr(self.c.spatial_samples, axis))
                self.assertEqual(line.shape, shape)
            for plane_name, plane in val.planes.items():
                monitor = self.c.models.eddy_diffusion.monitor_locations.planes[plane_name]
                axis = getattr(monitor, "axis")
                shape = [self.c.time_samples, 0, 0]
                for i, dim in enumerate(list(axis)):
                    shape[i + 1] = getattr(self.c.spatial_samples, dim)
                shape = tuple(shape)
                self.assertEqual(plane.shape, shape)


if __name__ == "__main__":
    unittest.main()



import unittest

import os

from ridt.config import ConfigFileParser


class ST05(unittest.TestCase):

    """System Test 05. Verify that the system can
       parse all plotting options from the JSON file,
       for point, line, plane monitor points."""

    def setUp(self) -> None:

        """setUp method that instantiates the
        :class:`~.RIDTConfig` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with ConfigFileParser(path) as cfp:
            self.c = cfp

    def test_verify(self):

        """Verifies that the :class:`~.ModelSettings` class
        contains the correct the attributes."""

        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion, "points_plots"), True)
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion, "lines_plots"), True)
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion, "planes_plots"), True)

    def test_point_plots(self):

        """Verifies that the :class:`~.PointPlots` class
        contains the correct the attributes."""

        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.points_plots, "output"), True)

    def test_line_plots(self):

        """Verifies that the :class:`~.LinePlots` class
        contains the correct the attributes."""

        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.lines_plots, "output"), True)

    def test_contour_plots(self):

        """Verifies that the :class:`~.ContourPlots` class
        contains the correct the attributes."""

        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.planes_plots, "output"), True)
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.planes_plots, "number_of_contours"), True)
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.planes_plots, "range"), True)
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.planes_plots, "scale"), True)
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.planes_plots.contours, "min"), True)
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.planes_plots.contours, "max"), True)


if __name__ == "__main__":
    unittest.main()



from os.path import join

from numpy import load

from ridt.base import ComputationalSpace
from ridt.base import Error

from ridt.config import RIDTConfig

from .batchdatastore import BatchDataStore
from .directoryagent import DirectoryAgent
from .datastore import DataStore


class DataStoreReader:
    """A class that attempts to read a data store from disk.

    This will only work if the folder structure generated when writing the data
    store to disk is un modified.

    When creating a new instance of this class, an instance of
    :class:`~.DataStore` or :class:`~.BatchDataStore` will be returned instead
    of :class:`~.DataStoreReader`.

    Attributes
    ----------
    directory: :obj:`str`
        The directory where the data store is located.
    
    settings : :class:`~.RIDTConfig`
        The settings for the run output stored in the data store.
    
    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.

    space : :class:`~.ComputationalSpace`
        The :class:`~.ComputationalSpace` instance corresponding to the
        :attr:`settings` attribute.

    data_store : :obj:`Union`[:class:`~.DataStore`, :class:`~.BatchDataStore`]
        The parsed data store.

    """

    def __new__(cls, *args, **kwargs):
        instance = super(DataStoreReader, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return instance.data_store

    def __init__(self, settings: RIDTConfig, directory: str, quantity: str):
        """The :class:`DataStoreReader` constructor.

        Parameters
        ----------
        directory: :obj:`str`
            The directory where the data store is located.
        
        settings : :class:`~.RIDTConfig`
            The settings for the run output stored in the data store.
        
        quantity: :obj:`str`
            The string id for the quantity stored in the data  store.

        """
        self.directory = directory
        self.settings = settings
        self.quantity = quantity
        restrict = {"models": self.settings.dispersion_model}
        self.space = ComputationalSpace(self.settings, restrict)
        self.read()

    @property
    def geometries(self):
        """:obj:`list` [:obj:`str`] : the list of geometries selected for
        evaluation in :attr:`settings`.

        """
        locations = self.settings.models.eddy_diffusion.monitor_locations
        return [g for g, e in locations.evaluate.items() if e]

    def read(self) -> None:
        """This method attempts to load the data store  .

        If :attr:`settings` is a batch settings object then it will loop over
        all elements in the computational space.

        Returns
        -------
        None


        """
        if self.space.zero:
            self.data_store = self.load(self.settings, self.directory)
        else:
            self.data_store = BatchDataStore()
            with DirectoryAgent(self.directory, self.space.shape) as da:
                for idx, setting in enumerate(self.space.space):
                    da.create_data_dir(idx)
                    self.data_store.add_run(setting)
                    self.data_store[setting] = self.load(setting, da.ddir)
    
    def load(self, setting: RIDTConfig, directory: str):
        """Reads in each numpy array and stores it in the new data store.

        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings for the run output stored in the data store.

        directory: :obj:`str`
            The directory where the data store is located.

        Returns
        -------
        :class:`~.DataStore`
            The parsed data store.

        Raises
        ------
        :class:`~.DataStoreParsingError`
            If unable to read a numpy binary file from disk.

        """
        rv = DataStore()
        locations = setting.models.eddy_diffusion.monitor_locations

        for geometry in self.geometries:
            for name in getattr(locations, geometry).keys():
                fname = name + ".npy"
                folder = join(directory, geometry, self.quantity)
                try:
                    with open(join(folder, fname), 'rb') as f:
                        rv.add(geometry, name, load(f))
                except OSError as e:
                    raise DataStoreParsingError(fname, folder, e)

        return rv
 

class DataStoreParsingError(Error):
    """The exception raised when an OSError is raised when parsing data arrays
    from disk

    """
    def __init__(self, file_name: str, directory: str, error):
        """The constructor for the :class:`DataStoreParsingError` class.

        """
        msg = f"When loading '{file_name}' from the directory '{directory}'"\
              f" the following error was encountered: {error}"
        super().__init__(msg)
    


from copy import deepcopy

from numpy import nanstd
from numpy import nanmean

from ridt.config import RIDTConfig
from ridt.config import Units

from ridt.container import Domain

from ridt.data import DataStore
from ridt.data import UncertaintyMask

from .resultcontainers import Maximum
from .resultcontainers import Exceedance
from .resultcontainers import PercentExceedance
from .resultcontainers import MaxPercentExceedance


class DataStoreAnalyser:
    """The Data Store Analyser class.

    Iterates through all selected geometries and computes all quantities
    such as the maximum value etc.

    Attributes
    ----------
    setting : :class:`~.RIDTConfig`
        The settings for the run in question.
    
    domain : :class:`~.Domain`
        The instance of :class:`~.Domain` corresponding to :attr:`setting`.

    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.

    data_store : :class:`~.DataStore`
        The data store to be analysed.
    
    thresholds : :obj:`list` [:obj:`float`]
        The threshold values corresponding to :attr:`quantity` defined in
        :attr:`setting`.
    
    units : :class:`~.Units`
        The instance of :class:`~.Units` corresponding to :attr:`setting`.
    
    maximum : :obj:`list` [:class:`~.Maximum`]
        The lists of :class:`~.Maximum` instances created.
    
    exceedance : :obj:`list` [:class:`~.Exceedance`]
        The lists of :class:`~.Exceedance` instances created.
    
    percent_exceedance: :obj:`list` [:class:`~.PercentExceedance`]
        The lists of :class:`~.PercentExceedance` instances created.
    
    max_percent_exceedance: :obj:`list` [:class:`~.MaxPercentExceedance`]
        The lists of :class:`~.MaxPercentExceedance` instances created.
    
    """
    def __init__(self,
                 setting: RIDTConfig,
                 data_store: DataStore,
                 quantity: str):
        """The :class`~.DataStoreAnalyser` class initialiser.
        
        Calls the :meth:`~.DataStoreAnalyser.evaluate` method.

        If flag is set in config file, will exclude values within 2m of all
        sources.

        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings for the run in question.

        data_store : :class:`~.DataStore`
            The data store to be analysed.

        quantity : :obj:`str`
            The string id for the quantity stored in the data  store.

       """

        self.setting = setting
        self.units = Units(setting)
        self.domain = Domain(self.setting)
        self.quantity = quantity
        self.thresholds = self.threshold_converter()
        self.data_store = data_store

        if self.setting.models.eddy_diffusion.analysis.exclude_uncertain_values:
            self.exclude_uncertain_values()

        self.maximum = list()
        self.exceedance = list()
        self.percent_exceedance = list()
        self.max_percent_exceedance = list()

        self.evaluate()

    @property
    def geometries(self):
        """:obj:`list` [:obj:`str`] : the list of geometries selected for
        evaluation in :attr:`setting`.

        """
        locations = self.setting.models.eddy_diffusion.monitor_locations
        return [g for g, e in locations.evaluate.items() if e]

    def threshold_converter(self):
        """Converts the threshold into SI units.

        Converts the thresholds in :attr:`setting` corresponding to
        :attr:`quantity` into SI units.

        Returns
        -------
        :obj:`list` [:obj:`float`]
            The list of threshold values in SI units. 
        """
        tld = [t.value for t in getattr(self.setting.thresholds, self.quantity)]
        return getattr(self.units, f"{self.quantity}_converter")(tld)

    def evaluate(self):
        """Loops over all selected geometries and computes various quantities.

        The following quantities are evaluated:

        The maximum value.

        The fastest time to all threshold values.

        The fastest time to the defined percentage of domain to all threshold
        values.

        The largest percentage that exceeds all threshold values.

        A corresponding :class:`~.ResultContainer` derived class instances
        is created to hold each result.

        Returns
        -------
        None

        """
        p = self.setting.models.eddy_diffusion.analysis.percentage_exceedance
        for geometry in self.geometries:
            for id in getattr(self.data_store, geometry):
                index, value = self.data_store.maximum(geometry, id)
                cargs = (geometry, id, self.quantity)
                self.maximum.append(Maximum(self.setting, *cargs, index, value))
        for t in self.thresholds:
            for geometry in self.geometries:
                for id in getattr(self.data_store, geometry):
                    cargs = (geometry, id)
                    index = self.data_store.exceeds(*cargs, t)
                    self.exceedance.append(
                        Exceedance(self.setting, *cargs, self.quantity, index, t))
                    index = self.data_store.percentage_exceeds(*cargs, t, p)
                    self.percent_exceedance.append(
                        PercentExceedance(self.setting, *cargs, self.quantity, index, t, p))
                    index, value = self.data_store.percentage_exceeds_max(geometry, id, t)
                    self.max_percent_exceedance.append(
                        MaxPercentExceedance(self.setting, *cargs, self.quantity, value, index, t))
    
    def exclude_uncertain_values(self):
        """Sets values within 2m of source to nan.

        Iterates through all domains in data store and sets values within 
        2m of any source to nan.

        Returns
        -------
        None

        """
        new_data_store = deepcopy(self.data_store)
        um = UncertaintyMask(self.setting)
        for geometry in self.geometries:
            for id in getattr(new_data_store, geometry):
                data = um.mask(geometry, id, new_data_store.get(geometry, id))
                new_data_store.add(geometry, id, data)
        self.data_store = new_data_store

    @property
    def time_to_well_mixed(self):
        """Evaluates the time for system to become 'well mixed'

        Computes the time it takes for the full domain normalised standard
        deviation to become <= 0.1.

        Returns
        -------
        Union[:obj:`float`, :obj:`None`] 
            If there exists a time when well mixed state is acheived that time
            is returned, else :obj:`None`.

        """
        for i in range(self.setting.time_samples):
            d = self.data_store.get("domain", "domain")[i, :, :, :]
            value = nanstd(d) / nanmean(d)
            if value <= 0.1:
                return self.domain.time[i]
        return None



from ridt.base import Error

from ridt.config import RIDTConfig

from .datastore import DataStore


class BatchDataStore:
    """The data store than contains other data stores.

    This behaves like a dictionary with type restrictions, keys are
    :class:`~.RIDTConfig` instances and values are :class:`~.DataStore`
    instances. 

    Attributes
    ----------
    data_store : :obj:`Dict`[:class:`~.RIDTConfig`, :class:`~.DataStore`]
        The dictionary containing the data stores.

    """

    def __init__(self):
        """The :class:`BatchDataStore` constructor.

        """
        self.store = dict()

    def add_run(self, setting: RIDTConfig) -> None:
        """Add a new empty data store for a new settings instance.

        If settings instance already in :attr:`store`, does nothing.

        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings for the run in question.
        
        Returns
        -------
        None

        """
        if setting not in self.store:
            self.store[setting] = DataStore()
        else:
            pass
    
    def values(self):
        """Interface for standard :obj:`dict` functionality.

        """
        return self.store.values()
    
    def keys(self):
        """Interface for standard :obj:`dict` functionality.

        """
        return self.store.keys()

    def items(self):
        """Interface for standard :obj:`dict` functionality.

        """
        return self.store.items()

    def __getitem__(self, setting: RIDTConfig) -> DataStore:
        """Interface for standard :obj:`dict` functionality, with type restrict.

        Raises
        ------
        :class:`BatchDataStoreIDError`
            If the requested settings instance is not in the store.

        """
        try:
            return self.store[setting]
        except KeyError as e:
            raise BatchDataStoreIDError(setting)

    def __setitem__(self, setting: RIDTConfig, value: DataStore):
        """Interface for standard :obj:`dict` functionality, with type restrict.

        Raises
        ------
        :class:`BatchDataStoreIDError`
            If the requested settings instance is not in the store.

        :class:`BatchDataStoreValueError`
            If `value` is not a :class:`~.DataStore` instance.

        """
        if not isinstance(value, DataStore):
            raise BatchDataStoreValueError()
        try:
            self.store[setting] = value
        except KeyError as e:
            raise BatchDataStoreIDError(setting)


class BatchDataStoreIDError(Error):
    """The exception raised when the data store is queries with settings object 
    it does not recognise.

    """
    def __init__(self, setting: RIDTConfig):
        """The constructor for the :class:`BatchDataStoreIDError` class.

        """
        msg = f"The data store does not contain any data for "\
            f"settings object {setting}."
        super().__init__(msg)


class BatchDataStoreValueError(Error):
    """The exception raised when an instance of some class other than DataStore
    is passed to __setitem__ 

    """
    def __init__(self):
        """The constructor for the :class:`BatchDataStoreValueError` class.

        """
        msg = f"The data store can only hold {DataStore} instances."
        super().__init__(msg)




from typing import Iterable

from .ridtconfig import RIDTConfig

R = 8.3145 # Universal Gas Constant
D = 1.292 # Density of air at room temperature kg.m-3


class Units:
    """The class that perfoms all unit conversions throughout RIDT.

    Attributes
    ----------
    time : :obj:`str`
        The string representation of the selected time units.

    mass : :obj:`str`
        The string representation of the selected mass units.
    
    concentration : :obj:`str`
        The string representation of the selected concentration units.

    exposure : :obj:`str`
        The string representation of the selected exposure units.

    space: :obj:`str`
        The string representation of the selected spatial units.
    
    concentration_si : :obj:`str`
        The string representation of the SI concentration units.

    exposure_si : :obj:`str`
        The string representation of the SI exposure units.
    
    temperature : :obj:`float`
        The temperature value.
    
    pressure : :obj:`float`
        The pressure value.
    
    molecular_weight : :obj:`float`
        The material's molecular weight value.
    
    nmol : obj:`float`
        The number of moles of gas per unit volume.

    """
    def __init__(self, setting: RIDTConfig):
        """The constructor for the :class:`Units` class.

        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings object containing the units selections and definitions. 

        """
        self.time = setting.time_units
        self.mass = setting.mass_units
        self.concentration = setting.concentration_units
        self.exposure = setting.exposure_units
        self.space = setting.spatial_units
        self.concentration_si = "kg.m-3"
        self.exposure_si = "kg.s.m-3"
        self.temperature = setting.physical_properties.temperature
        self.pressure = setting.physical_properties.pressure
        self.molecular_weight = setting.physical_properties.agent_molecular_weight
        self.air_density = setting.physical_properties.air_density
        self.nmol = self.pressure / (self.temperature * R)

    @property
    def concentration_factor(self) -> float:
        """The conversion factor for provided concentration into SI units.

        This function will compute the scaling factor for the concentration
        quantities to and from SI units. This is because all calculations RIDT
        performs internally are in SI units.

        Returns
        -------
        :obj:`float`
            The unit conversion factor.

        Raises
        ------
        :obj:`ValueError`
            If the concentration units provided are not valid.

        """
        if self.concentration == "kg.m-3":
            return 1.0
        elif self.concentration == "kg.kg-1":
            return 1 * self.air_density
        elif self.concentration == "mg.m-3":
            return 1e-6
        elif self.concentration == "ppm":
            return (self.molecular_weight * self.nmol) * 1e-6
        elif self.concentration == "ppb":
            return (self.molecular_weight * self.nmol) * 1e-9
        elif self.concentration == "ppt":
            return (self.molecular_weight * self.nmol) * 1e-12
        else:
            raise ValueError(f"{self.concentration} not a valid unit")

    @property
    def exposure_factor(self):
        """The conversion factor for provided exposure into SI units.

        This function will compute the scaling factor for the exposure
        quantities to and from SI units. This is because all calculations RIDT
        performs internally are in SI units.

        Returns
        -------
        :obj:`float`
            The unit conversion factor.

        Raises
        ------
        :obj:`ValueError`
            If the exposure units provided are not valid.

        """
        if self.exposure == "kg.s.m-3":
            return 1.0
        elif self.exposure == "mg.min.m-3":
            return 1e-6 * 60
        else:
            raise ValueError(f"{self.exposure} not a valid unit")

    def concentration_converter(self, values: Iterable[float]):
        """Converts all concentration values in iterable to SI units.

        Parameters
        ----------
        values : :obj:`Iterable`[:obj:`float`]
            The iterable of values to be converted.

        Returns
        -------
        :obj:`Iterable`[:obj:`float`]
            A new iterable containing the scaled values.
        """
        return [v * self.concentration_factor for v in values]

    def exposure_converter(self, values: Iterable):
        """Converts all exposure values in iterable to SI units.

        Parameters
        ----------
        values : :obj:`Iterable`[:obj:`float`]
            The iterable of values to be converted.

        Returns
        -------
        :obj:`Iterable`[:obj:`float`]
            A new iterable containing the scaled values.
        """
        return [v * self.exposure_factor for v in values]



import unittest
from os import listdir
from os.path import join
from os import remove
import shutil

from ridt.config import ConfigFileParser

from ridt.container import Domain
from ridt.container.eddydiffusionrun import EddyDiffusionRun

from ridt.data import BatchDataStore
from ridt.data import DataStoreReader


class ST25(unittest.TestCase):

    """System Test 25. Test the system
       is able to output the data store
       created during a batch run or
       single run mode to disk."""

    def setUp(self) -> None:
        pass

    def tearDown(self) -> None:
        pass

    def test_verify(self):
        pass


if __name__ == "__main__":
    unittest.main()



from ridt.base import Settings
from ridt.base import Terminus
from ridt.base import List
from ridt.base import Dict
from ridt.base import Number
from ridt.base import StringSelection
from ridt.base import ConsistencyError

from numpy import min

import warnings


def custom_formatwarning(msg, *args, **kwargs):
    return "Warning: " + str(msg) + '\n'

warnings.formatwarning = custom_formatwarning


class RIDTConfig(Settings):
    """The settings object for ridt.

    This takes a dictionary and checks that all relevant information exists
    in the dictionary.

    Attributes
    ----------
    well_mixed : :obj:`bool`
        The flag indicating whether to evaluate the well mixed model.

    eddy_diffusion: :obj:`bool`
        The flag indicating whether to evaluate the eddy diffusion model.
    
    compute_exposure: :obj:`bool`
        The flag indicating whether to compute and perfom analysis on the
        exposure.

    write_data_to_csv: :obj:`bool`
        The flag indicating whether to write all computed data to csv files, in
        addition to the numpy bindary arrays (*.npy).

    time_units : :class:`~.TimeUnits`
        The temporal units selection.
    
    time_samples : :class:`~.NonNegativeInteger`
        The temporal discretisation.
    
    total_time : :class:`~.NonNegativeFloat`
        The total time for the simulation.
    
    dimensions : :class:`~.Dimensions`
        The spatial bounds of the system.
    
    spatial_samples : :class:`~.SpatialSamples`
        The spatial discretisation of the system.

    spatial_units : :class:`~.SpatialUnits`
        The spatial units selection.

    physical_properties : :class:`~.PhysicalProperties`
        The physical properties of the system.
    
    concentration_units : :class:`~.ConcentrationUnits`
        The concentration units selection.
    
    exposure_units : :class:`~.ExposureUnits`
        The exposure units selection.
    
    mass_units : :class:`~.MassUnits`
        The mass units selection.
    
    fresh_air_flow_rate : :class:`~.FreshAirFlowRate`
        The fresh air flow rate.

    fresh_air_flow_rate_units : :class:`~.FreshAirFlowRateUnits`
        The fresh air flow rate unit selection.
    
    modes : :class:`~.Modes`
        Source settings.
    
    thresholds : :class:`~.Thresholds`
        Threshold settings.
    
    models : :class:`~.Models`
        Model specific settings.

    Parameters
    ----------
    value : :obj:`dict`
        The dictionary of values to be verified and stored.

    Raises
    ------
    
    """

    @Settings.assign
    def __init__(self, values: dict):
        """The RIDTConfig initaliser

        Parameters
        ----------
        value : :obj:`dict`
            The dictionary of values to be verified and stored.


        """
        self.ridt_version = RIDTVersion
        self.well_mixed = bool
        self.eddy_diffusion = bool
        self.compute_exposure = bool
        self.write_data_to_csv = bool

        self.integration_method = IntegrationMethod

        self.time_units = TimeUnits
        self.time_samples = TimeSamples
        self.total_time = NonNegativeFloat

        self.dimensions = Dimensions
        self.spatial_samples = SpatialSamples

        self.spatial_units = SpatialUnits
        self.physical_properties = PhysicalProperties
        self.concentration_units = ConcentrationUnits
        self.exposure_units = ExposureUnits
        self.mass_units = MassUnits

        self.fresh_air_flow_rate_units = FreshAirFlowRateUnits
        self.fresh_air_flow_rate = NonNegativeFloat 

        self.modes = ModeSettings
        self.thresholds = Thresholds

        self.models = ModelSettings

    def consistency_check(self):
        """Verifies the self consistency of the settings object.

        Returns
        -------
        None

        Raises
        ------
        :class:`~.ConsistencyError`
            If any sources lie outside the simulation domain.

        :class:`~.ConsistencyError`
            If any sources' release times are outside the simulation domain.

        :class:`~.ConsistencyError`
            If any monitor locations are outside of the simulation domain.

        :class:`~.ConsistencyError`
            If more than 5 thresholds of a given type are defined.

        :class:`~.ConsistencyError`
            If the number of a given type of plots requested is greater than
            the time discretisation.

        """

        for mode in ["instantaneous", "infinite_duration", "fixed_duration"]:
            for key, value in getattr(self.modes, mode).sources.items():
                dim = self.dimensions
                for axis in ["x", "y", "z"]:
                    par = getattr(value, axis)
                    bound = min(getattr(dim, axis))
                    if isinstance(par, list):
                        for item in par:
                            if item < 0 or item > bound:
                                raise ConsistencyError(
                                    f"{mode} source {key} x position is "
                                    f"outside space domain (0, {bound}).")
                    else:
                        if par < 0 or par > bound:
                            raise ConsistencyError(
                                f"{mode} source {key} x position is "
                                f"outside space domain (0, {bound}).")

        for mode in ["instantaneous", "infinite_duration"]:
            for key, value in getattr(self.modes, mode).sources.items():
                if isinstance(value.time, list):
                    for item in value.time:
                        if item > self.total_time:
                            raise ConsistencyError(
                                f"{mode} source {key} time is "
                                f"outside time domain [0, {self.total_time}).")
                else:
                    if value.time > self.total_time:
                        raise ConsistencyError(
                            f"{mode} source {key} time is "
                            f"outside time domain [0, {self.total_time}).")

        for key, value in getattr(self.modes, "fixed_duration").sources.items():
            if isinstance(value.start_time, list):
                for item in value.start_time:
                    if item > self.total_time:
                        raise ConsistencyError(
                            f"{mode} source {key} start time is "
                            f"outside time domain [0, {self.total_time}).")
            else:
                if value.start_time > self.total_time:
                    raise ConsistencyError(
                        f"{mode} source {key} start time is "
                        f"outside time domain [0, {self.total_time}).")

            if isinstance(value.end_time, list):
                for item in value.end_time:
                    if item > self.total_time:
                        raise ConsistencyError(
                            f"{mode} source {key} end time is "
                            f"outside time domain [0, {self.total_time}).")
            else:
                if value.end_time > self.total_time:
                    raise ConsistencyError(
                        f"{mode} source {key} end time is "
                        f"outside time domain [0, {self.total_time}).")
        
        dims = ["x", "y", "z"]
        for plane in self.models.eddy_diffusion.monitor_locations.planes.values():
            dim = [axis for axis in dims if axis not in str(plane.axis)][0]
            if isinstance(plane.distance, list):
                if max(plane.distance) > min(getattr(self.dimensions, dim)):
                    raise ConsistencyError(f"{plane} is outside the space domain")
            else:
                if plane.distance > min(getattr(self.dimensions, dim)):
                    raise ConsistencyError(f"{plane} is outside the space domain")

        for key, point in self.models.eddy_diffusion.monitor_locations.points.items():
            for dim in dims:
                if getattr(point, dim) > min(getattr(self.dimensions, dim)):
                    raise ConsistencyError(
                        f"{key}'s {dim} value, is outside space domain "
                        f"(0, {getattr(self.dimensions, dim)})")

        for key, line in self.models.eddy_diffusion.monitor_locations.lines.items():
            for dim in dims:
                if getattr(line.point, dim) > min(getattr(self.dimensions, dim)):
                    raise ConsistencyError(
                        f"{key}'s {dim} value, is outside space domain "
                        f"(0, {getattr(self.dimensions, dim)})")

        thresh = self.thresholds
        if len(thresh.concentration) > 5 or len(thresh.exposure) > 5:
            raise ConsistencyError(f"Cannot exceed more than 5 thresholds")
 
        line_number = self.models.eddy_diffusion.lines_plots.number
        if isinstance(line_number, list):
            for item in line_number:
                if item > self.time_samples:
                    raise ConsistencyError(
                        f"The number of requested line plots ({line_number}) cannot exceed the "
                        f"number of time samples ({self.time_samples}).")
        else:
            if line_number > self.time_samples:
                raise ConsistencyError(
                    f"The number of requested line plots ({line_number}) cannot exceed the "
                    f"number of time samples ({self.time_samples}).")
 
        contour_number = self.models.eddy_diffusion.planes_plots.number
        if isinstance(contour_number, list):
            for item in contour_number:
                if item > self.time_samples:
                    raise ConsistencyError(
                        f"The number of requested contour plots ({item}) cannot exceed the "
                        f"number of time samples ({self.time_samples}.)")
        else:
            if contour_number > self.time_samples:
                raise ConsistencyError(
                    f"The number of requested contour plots ({contour_number}) cannot exceed the "
                    f"number of time samples ({self.time_samples}).")


class RIDTVersion(StringSelection):
    """The setting that indicates the version of ridt for the config file

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The RIDTVersion class initialiser

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "v1.0"
        ]
    
    def check(self):
        pass
    

class TimeSamples(Number):
    """The :class:`~.TimeSamples` class. It inherits from
    :class:`~.Number`.

    """
    @Terminus.assign
    def __init__(self, value: int):
        """The constructor for the :class:`~.TimeSamples` class.

        Parameters
        ----------
        value : :obj:`int`
            The float value that is being checked.

        """
        self.type = int 

    def check(self):
        """Abstract method from :class:`~.Terminus`.
        Raises
        ------
        ValueError
            If the value isn't greater than 0.
        """
        self.lower_bound_exclusive(1)


class PhysicalProperties(Settings):
    """The physical properties settings object.

    Attributes
    ----------
    agent_molecular_weight_units : :class:`~.AgentMolecularWeightUnits`
        The agent molecular weight units selection.

    agent_molecular_weight : :class:`~.AgentMolecularWeight`
        The agent molecular weight.
    
    pressure_units : :class:`~.PressureUnits`
        The pressure units selection.

    pressure: :class:`~.Pressure`
        The pressure of the atmosphere.
    
    temperature_units : :class:`~.TemperatureUnits`
        The temperature units selection.

    temperature: :class:`~.Temperature`
        The temperature of the atmosphere.
    
    air_density_units: :class:`~.AirDensityUnits`
        The air density units
    
    air_density: :class:`~.AirDensity`
        The air density.

    """
    @Settings.assign
    def __init__(self, values: dict):
        """The PhysicalProperties initialiser

        Parameters
        ----------
        value : :obj:`dict`
            The dictionary of values to be verified and stored.

        """
        self.agent_molecular_weight_units = AgentMolecularWeightUnits
        self.agent_molecular_weight = PositiveFloat 
        self.pressure_units = PressureUnits
        self.pressure = PositiveFloat 
        self.temperature_units = TemperatureUnits
        self.temperature = PositiveFloat
        self.air_density_units = AirDensityUnits
        self.air_density = PositiveFloat 


class AgentMolecularWeightUnits(StringSelection):
    """The agent molecular weights units selection class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The AgentMolecularWeightUnits class initialiser

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "kg.mol-1"
        ]
    
    def check(self):
        pass
    

class IntegrationMethod(StringSelection):
    """The integration methods selection class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The IntegrationMethod class initialiser

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "cumulativetrapezoidal",
            "romberg"
        ]
    
    def check(self):
        pass
 

class PressureUnits(StringSelection):
    """The pressure units selection class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The PressureUnits class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "Pa"
        ]
    
    def check(self):
        pass


class TemperatureUnits(StringSelection):
    """The temperature units selection class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The TemperatureUnits class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "K"
        ]
    
    def check(self):
        pass

class AirDensityUnits(StringSelection):
    """The air density units selection class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The AirDensityUnits class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "kg.m-3"
        ]
    
    def check(self):
        pass



class FreshAirFlowRate(Number):
    """The FreshAirFlowRate class

    Stores the value of the fresh air flow rate.

    """

    @Terminus.assign
    def __init__(self, value: float):
        self.type = float
    
    def check(self):
        """Checks that the value is in the required bounds.

        """
        self.lower_bound(5e-4)
        self.upper_bound(5e2)


class DispersionModel(StringSelection):
    """The dispersion model selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The DispersionModel class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "eddy_diffusion",
            "well_mixed"
        ]

    def check(self):
        pass


class ReleaseType(StringSelection):
    """The release type selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The ReleaseType class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "instantaneous",
            "infinite_duration",
            "fixed_duration"
        ]

    def check(self):
        pass


class TimeUnits(StringSelection):
    """The time units selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The TimeUnits class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "s",
        ]

    def check(self):
        pass


class ConcentrationUnits(StringSelection):
    """The time units selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The ConcentrationUnits class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "kg.m-3",
            "kg.kg-1",
            "mg.m-3",
            "ppm",
            "ppb",
            "ppt"
        ]

    def check(self):
        pass


class ExposureUnits(StringSelection):
    """The time units selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The ExposureUnits class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "mg.min.m-3",
            "kg.s.m-3"
        ]

    def check(self):
        pass


class SpatialUnits(StringSelection):
    """The time units selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The SpatialUnits class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "m",
        ]

    def check(self):
        pass


class MassUnits(StringSelection):
    """The mass units selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The MassUnits class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "kg",
        ]

    def check(self):
        pass


class FreshAirFlowRateUnits(StringSelection):
    """The fresh air flow rate units selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The FreshAirFlowRateUnits class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "m3.s-1",
        ]

    def check(self):
        pass


class ModelSettings(Settings):
    """The Model Settings class.

    Attributes
    ----------
    eddy_diffusion: :class:`~.EddyDiffusion`
        The  eddy diffusion model settings.


    """
    @Settings.assign
    def __init__(self, values: dict):
        """The ModelSettings class initialiser.

        Parameters
        ----------
        values : :obj:`dict`
            The dictionary corresponding to the instantaneous
            settings configurations.
        """
        self.eddy_diffusion = EddyDiffusion


class ModeSettings(Settings):
    """The :class:`~.ModeSettings` class. It inherits from :class:`~.Settings`
    class

    Attributes
    ----------
    instantaneous: :class:`~.InstantaneousSettings`
        A :class:`~.Settings` child that contains paths
        to the Instantaneous configurations.

    infinite_duration: :class:`~.InfiniteDurationSettings`
        A :class:`~.Settings` child that contains paths
        to the InfiniteDuration configurations.

    fixed_duration: :class:`~.FixedDurationSettings`
        A :class:`~.Settings` child that contains paths
        to the FixedDuration configurations.

    """
    @Settings.assign
    def __init__(self, values: dict):
        """The ModeSettings class initialiser.

        Parameters
        ----------
        values : :obj:`dict`
            The dictionary corresponding to the mode settings
            configurations.
        """
        self.instantaneous = InstantaneousSettings
        self.infinite_duration = InfiniteDurationSettings
        self.fixed_duration = FixedDurationSettings


class InstantaneousSettings(Settings):

    """The :class:`~.InstantaneousSettings` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    sources: :class:`~.Dict`
        Path to the dictionary of instantaneous sources within
        the simulation.
    """

    @Settings.assign
    def __init__(self, values: dict):
        """The InstantaneousSettings class initialiser.

        Parameters
        ----------
        values : :obj:`dict`
            The dictionary corresponding to the instantaneous
            settings configurations.
        """

        self.sources = InstantaneousSourceDict


class InstantaneousSourceDict(Dict):
    """The :class:`~.InstantaneousSourceDict` class. It inherits from
    :class:`~.Dict`.

    Attributes
    ---------
    type: :class:`~.InstantaneousSource`
        Path to the dictionary of values of the instantaneous
        sources.
    """
    @Dict.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.InstantaneousSourceDict` class.

        Parameters
        ----------
        values : :obj:`dict`
            The value corresponding to the instantaneous
            source configurations.
        """
        self.type = InstantaneousSource


class InstantaneousSource(Settings):
    """The :class:`~.InstantaneousSource` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    x: :class:`~.NonNegativeFloat`
        The x coordinate of the source.

    y: :class:`~.NonNegativeFloat`
        The y coordinate of the source.

    z: :class:`~.NonNegativeFloat`
        The z coordinate of the source.

    mass: :class:`~.NonNegativeFloat`
        The mass of the material released.

    time: :class:`~.NonNegativeFloat`
        The time that the material is released.

    """
    @Settings.assign
    def __init__(self, values: dict):
        """The InstantaneousSource class initialiser.

        Parameters
        ----------
        values : :obj:`dict`
            The dictionary corresponding to the instantaneous
            sources.
        """
        self.x = NonNegativeFloat
        self.y = NonNegativeFloat
        self.z = NonNegativeFloat
        self.mass = NonNegativeFloat
        self.time = NonNegativeFloat


class InstantaneousReleaseMass(Number):
    """The instantaous release mass setting.

    Container the value of an instantaneous release mass.

    """
    @Terminus.assign
    def __init__(self, value: float):
        self.type = float

    def check(self):
        self.lower_bound(1e-6)
        self.upper_bound(10.0)


class ReleaseRate(Number):
    """The release rate setting.

    Contains the value for the rate of release for a source.
    
    """
    @Terminus.assign
    def __init__(self, value: float):
        self.type = float

    def check(self):
        self.lower_bound(1e-6)
        self.upper_bound(10.0)


class InfiniteDurationSettings(Settings):
    """The :class:`~.InfiniteDurationSettings` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    sources: :class:`~.Dict`
        Path to the dictionary of infinite duration sources within
        the simulation.

    """
    @Settings.assign
    def __init__(self, values: dict):
        """The InfiniteDurationSettings class initialiser.

        Parameters
        ----------
        values : :obj:`dict`
            The dictionary corresponding to the infinite duration
            sources.

        """
        self.sources = InfiniteDurationSourceDict


class InfiniteDurationSourceDict(Dict):
    """The :class:`~.InfiniteDurationSourceDict` class. It inherits from
     :class:`~.Dict`.

     Attributes
     ---------
     type: :class:`~.InfiniteDurationSource`
         Path to the dictionary of values of the infinite duration
         sources.

     """
    @Dict.assign
    def __init__(self, values: dict):
        """The InfiniteDurationSourceDict  class initialiser.

        Parameters
        ----------
        values : :obj:`dict`
            The value corresponding to the infinite duration
            sources.
        """
        self.type = InfiniteDurationSource


class InfiniteDurationSource(Settings):
    """The :class:`~.InfiniteDurationSource` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    x: :class:`~.NonNegativeFloat`
        The x coordinate of the source.

    y: :class:`~.NonNegativeFloat`
        The y coordinate of the source.

    z: :class:`~.NonNegativeFloat`
        The z coordinate of the source.

    rate: :class:`~.NonNegativeFloat`
        The mass of the material released.

    time: :class:`~.NonNegativeFloat`
        The time that the material is released.

    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.InfiniteDurationSource` class.

        Parameters
        ----------
        values : :obj:`dict`
            The dictionary corresponding to the infinite duration
            sources configurations.

        """
        self.x = NonNegativeFloat
        self.y = NonNegativeFloat
        self.z = NonNegativeFloat
        self.rate = NonNegativeFloat
        self.time = NonNegativeFloat


class FixedDurationSettings(Settings):
    """The :class:`~.FixedDurationSettings` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    sources: :class:`~.Dict`
        Path to the dictionary of fixed duration sources within
        the simulation.
    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.FixedDurationSettings` class.

        Parameters
        ----------
        values : :obj:`dict`
            The dictionary corresponding to the fixed duration sources.
        """
        self.sources = FixedDurationSourceDict


class FixedDurationSourceDict(Dict):

    """The :class:`~.FixedDurationSourceDict` class. It inherits from
    :class:`~.Dict`.

    Attributes
    ---------
    type: :class:`~.FixedDurationSource`
        Path to the dictionary of values of the fixed duration
        sources.

    """

    @Dict.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.FixedDurationSourceDict` class.

        Parameters
        ----------
        values : :obj:`dict`
            The value corresponding to the fixed duration sources.

        """
        self.type = FixedDurationSource


class FixedDurationSource(Settings):
    """The :class:`~.FixedDurationSource` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    x: :class:`~.NonNegativeFloat`
        The x coordinate of the source.

    y: :class:`~.NonNegativeFloat`
        The y coordinate of the source.

    z: :class:`~.NonNegativeFloat`
        The z coordinate of the source.

    rate: :class:`~.NonNegativeFloat`
        The mass of the material released.

    start_time: :class:`~.NonNegativeFloat`
        The time that the material is released.

    end_time: :class:`~.NonNegativeFloat`
        The time that the source stops emitting
        the material.

    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.FixedDurationSource` class.

        Parameters
        ----------
        values : :obj:`dict`
            The dictionary corresponding to the fixed duration
             sources configurations.

        """
        self.x = NonNegativeFloat
        self.y = NonNegativeFloat
        self.z = NonNegativeFloat
        self.rate = NonNegativeFloat
        self.start_time = NonNegativeFloat
        self.end_time = NonNegativeFloat


class Point(Settings):
    """The :class:`~.Point` class. It inherits from :class:`~.Settings`.

    Attributes
    ---------
    x: :class:`~.NonNegativeFloat`
        The x coordinate of the point.

    y: :class:`~.NonNegativeFloat`
        The y coordinate of the point.

    z: :class:`~.NonNegativeFloat`
        The z coordinate of the point.

    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.Point` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the point location.

        """
        self.x = NonNegativeFloat
        self.y = NonNegativeFloat
        self.z = NonNegativeFloat


class Thresholds(Settings):
    """The :class:`~.Thresholds` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    concentration: :class:`~.ThresholdList`
        Path to the list of concentration thresholds within
        the simulation.

    exposure: :class:`~.ThresholdList`
        Path to the list of exposure thresholds within
        the simulation.
    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.Thresholds` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the exposure
            and concentration thresholds.

        """
        self.concentration = ThresholdList
        self.exposure = ThresholdList


class ThresholdList(List):
    """The :class:`~.ThresholdList` class. It inherits from
    :class:`~.List`.

    Attributes
    ---------
    type: :class:`~.Threshold`
        Path to the dictionary of values of the thresholds
        in the simulation.

    """
    @List.assign
    def __init__(self, values: list):
        """The constructor for the :class:`~.ThresholdList` class.

        Parameters
        ----------
        values : :obj:`list`
            The values corresponding to the thresholds.

        """
        self.type = NonNegativeFloat


class EddyDiffusion(Settings):
    """The :class:`~.EddyDiffusion` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    dimensions: :class:`~.Dimensions`
        Path to the dimensions of the container.

    spatial_samples: :class:`~.SpatialSamples`
        Path to the spatial_samples container.

    coefficient: :class:`~.Coefficient`
        Path to the eddy diffusion
        coefficient calculations.

    images: :class:`~.Images`
        Path to the image configurations.

    planes_plots: :class:`~.ContourPlots`
        Path to the contour plots configurations.
    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.EddyDiffusion` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the eddy diffusion
            model type.
        """
        self.monitor_locations = MonitorLocations
        self.coefficient = Coefficient
        self.images = Images
        self.analysis = AnalysisSettings
        self.planes_plots = ContourPlots
        self.lines_plots = LinePlots
        self.points_plots = PointPlots


class AnalysisSettings(Settings):
    """The analysis settings 

    Attributes
    ----------
    perform_analysis: :obj:`bool`
        Whether perform analysis or not.

    percentage_exceedance : :class:`~.Percentage`
        The percentage which must be exceeded in percent exceed threshold
        calcluations.
    
    exclude_uncertain_values: :obj:`bool`
        Whether to exclude values within a sphere of a given radius centered on
        all sources from analysis. 
    
    exclude_radius_meters: :obj:`bool`
        The radius of the sphere, inside which to exclude values.

    """
    @Settings.assign
    def __init__(self, values: dict):
        """The AnalysisSettings class initialiser.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the eddy diffusion
            model type.

        """
 
        self.perform_analysis = bool
        self.percentage_exceedance = Percentage
        self.exclude_uncertain_values = bool
        self.exclude_radius_meters = PositiveFloat


class LinePlots(Settings):
    """The :class:`~.LinePlots` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ----------
    output : :obj:`bool`
        Whether or not we output line plots.

    """
    @Settings.assign
    def __init__(self, values: dict):
        self.output = bool
        self.animate = bool
        self.scale = ScaleType
        self.number = PositiveInt 


class PointPlots(Settings):
    """The :class:`~.PointPlots` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ----------
    output : :obj:`bool`
        Whether or not we output point plots.

    """
    @Settings.assign
    def __init__(self, values: dict):
        self.time_axis_units = TimeAxisUnits
        self.output = bool
        self.scale = ScaleType


class TimeAxisUnits(StringSelection):
    """The time axis unit selection

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`TimeAxisUnits` class.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "s",
            "m",
            "h"
        ]

    def check(self):
        pass



class MonitorLocations(Settings):
    """The :class:`~.MonitorLocations` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    points : :class:`~.Points`
        The dictionary of points to be monitored.

    lines : :class:`~.Lines`
        The dictionary of lines to be monitored.

    planes : :class:`~.Planes`
        The dictionary of planes to be monitored.

    """
    @Settings.assign
    def __init__(self, values: dict):
        self.evaluate = EvaluateFlags
        self.points = Points
        self.lines = Lines
        self.planes = Planes
        self.domain = FullDomain
    
    def consistency_check(self):
        for item in self.evaluate.keys():
            if item not in self.__dict__:
                raise ConsistencyError(f"{item} is not a type if monitor location")


class EvaluateFlags(Dict):

    @Settings.assign
    def __init__(self, values: dict):
        self.type = bool


class Dimensions(Settings):
    """The :class:`~.Dimensions` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    x: :class:`~.NonNegativeFloat`
        The x magnitude of the container.
    y: :class:`~.NonNegativeFloat`
        The y magnitude of the container.
    z: :class:`~.NonNegativeFloat`
        The z magnitude of the container.
    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.Dimensions` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the dimensions
            of the container.
        """
        self.x = PositiveFloat 
        self.y = PositiveFloat 
        self.z = PositiveFloat


class Length(Number):
    @Terminus.assign
    def __init__(self, value: float):
        self.type = float

    def check(self):
        self.lower_bound(2)
        self.upper_bound(100)


class Width(Number):
    @Terminus.assign
    def __init__(self, value: float):
        self.type = float

    def check(self):
        self.lower_bound(2)
        self.upper_bound(100)


class Height(Number):
    @Terminus.assign
    def __init__(self, value: float):
        self.type = float

    def check(self):
        self.lower_bound(2)
        self.upper_bound(20)


class SpatialSamples(Settings):
    """The :class:`~.SpatialSamples` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    x: :class:`~.NonNegativeInteger`
        The x magnitude of the container.
    y: :class:`~.NonNegativeInteger`
        The y magnitude of the container.
    z: :class:`~.NonNegativeInteger`
        The z magnitude of the container.
    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.SpatialSamples` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the spatial samples
            of the container.
        """
        self.x = SpatialDiscretisation 
        self.y = SpatialDiscretisation
        self.z = SpatialDiscretisation


class SpatialDiscretisation(Number):
    @Terminus.assign
    def __init__(self, value: int):
        self.type = int
    
    def check(self):
        self.lower_bound(2)


class Coefficient(Settings):
    """The :class:`~.Coefficient` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    calculation: :obj:`str`
        A string dictating the way the eddy diffusion 
        coefficient will be calculated.
    value: :obj:`float`
        The explicit value of the eddy diffusion coefficient.
    tkeb: :class:`~.TKEB`
        The path leading to the TKEB configurations.
    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.Coefficient` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the coefficient
            calculations.
        """
        self.calculation = CoefficientMode
        self.value = ExplicitCoefficient
        self.tkeb = TKEB


class CoefficientMode(StringSelection):
    """The diffusion coefficient mode selection

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`CoefficientMode` class.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "EXPLICIT",
            "TKEB"
        ]

    def check(self):
        pass


class ExplicitCoefficient(Number):
    """The setting for the explicit value of the diffusion coefficient.

    Inherits from :class:`~.Number`.
    """
    @Terminus.assign
    def __init__(self, value: float):
        """The constructor for the :class:`~.ExplicitCoefficient` class.

        Parameters
        ----------
        value : :obj:`float`
            The float value that is being checked.
        """
        self.type = float

    def check(self):
        """Abstract method from :class:`~.Terminus`.

        """
        if self.value < 1e-3:
            warnings.warn("The diffusion coefficient < 1e-3.")
        if self.value > 1.0:
            warnings.warn("The diffusion coefficient > 1.0.")


class TKEB(Settings):
    """The :class:`~.TKEB` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    number_of_supply_vents: :obj:`int`
        The number of supply vents in the system.
    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.TKEB` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the TKEB
            equation.
        """
        self.bound = TKEBBound
        self.total_air_flow_rate = TotalAirFlowRate 
        self.number_of_supply_vents = NumberOfSupplyVents 


class TKEBBound(StringSelection):
    """The TKEB bound selection class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.
    
    """
    @Terminus.assign
    def __init__(self, values: dict):
        """The TKEBBound class initialiser.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.
        
        """
        self.options = [
            "lower",
            "regression",
            "upper"
        ]

    def check(self):
        pass


class TotalAirFlowRate(Number):
    """Total air flow rate settings

    Contains the value of the total air flow rate.

    """
    @Terminus.assign
    def __init__(self, value: float):
        self.type = float

    def check(self):
        self.lower_bound(5e-4)
        self.upper_bound(5e2)


class NumberOfSupplyVents(Number):
    """Number of supply vents.

    Contains the number of supply vents.

    """
    @Terminus.assign
    def __init__(self, value: int):
        self.type = int 

    def check(self):
        self.lower_bound(1)
        self.upper_bound(50)


class Images(Settings):
    """The :class:`~.Images` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    quantity: :obj:`int`
        The number of images.
    max_error: :class:`~.Percentage`
        The max error allowed when changing the number of image sources by 1.
    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.TKEB` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the TKEB
            equation.
        """
        self.mode = ImageMode
        self.quantity = ImageSourceNumber 


class ImageSourceNumber(Number):
    @Terminus.assign
    def __init__(self, value: int):
        self.type = int
    
    def check(self):
        self.lower_bound(0)
        self.upper_bound(20)


class ImageMode(StringSelection):
    """The image mode selection

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`ImageMode` class.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "manual",
            "auto"
        ]

    def check(self):
        pass


class ContourPlots(Settings):
    """The :class:`~.ContourPlots` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    output : :obj:`bool`
        Whether or not we output the contour plots.

    creation_frequency: :class:`~.NonNegativeFloat`
        How often we wish to create a contour plot of each plane in 1/s.

    number_of_contours: :obj:`int`
        The number of contours depicted.

    range: :obj:`str`
        The range of the contours. Can be auto
        or manually set.

    scale: :obj:`str`
        The scale of the contours. Either logarithmic
        or linear.

    contours: :class:`~.ManualContours`
        If range is manual. Then set the min and max
        values of the contours.

    """
    @Settings.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.ContourPlots` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the contour plots.
        """
        self.output = bool
        self.animate = bool
        self.number = PositiveInt 
        self.number_of_contours = NonNegativeInteger
        self.range = RangeMode
        self.scale = ScaleType 
        self.contours = ManualContours
    
    def consistency_check(self):
        if self.contours.min >= self.contours.max:
            raise ConsistencyError(
                f"contour min ({self.contours.min})"
                f" >= contour max ({self.contours.max})")


class RangeMode(StringSelection):
    """The contour range mode selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The constructor for the :class:`RangeMode` class.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "auto",
            "manual"
        ]

    def check(self):
        pass


class ScaleType(StringSelection):
    """The contour scale type selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`ScaleType` class.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "linear",
            "logarithmic"
        ]

    def check(self):
        pass


class Points(Dict):

    """The :class:`~.Points` class. It inherits from
    :class:`~.Dict`.

    Attributes
    ---------
    type: :class:`~.Point`
        Path to the dictionary of values for the points to be monitored

    """
    @Dict.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.Points` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the points to be monitored.

        """
        self.type = Point


class Lines(Dict):

    """The :class:`~.Lines` class. It inherits from
    :class:`~.Dict`.

    Attributes
    ---------
    type: :class:`~.Line`
        Path to the dictionary of lines to be monitored.

    """
    @Dict.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.Lines` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the lines to be monitored.

        """
        self.type = Line


class Line(Settings):
    """The :class:`~.Line` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    start_point: :class:`~.Point`
        The location of the start of the line in 3D space.

    length: :class:`~.Point`
        The length of the line in 3D space.

    direction: :obj:`str`
        The axis in which the line extends.

    """
    @Settings.assign
    def __init__(self, value: dict):
        """The constructor for the :class:`~.Line` class.

        Parameters
        ----------
        value : :obj:`dict`
            The values corresponding to each line.
        """
        self.point = Point
        self.parallel_axis = ParallelAxis


class ParallelAxis(StringSelection):
    """The parallel axis selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The constructor for the :class:`ParallelAxis` class.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "x",
            "y",
            "z"
        ]

    def check(self):
        pass


class Planes(Dict):
    """The :class:`~.Planes` class. It inherits from
    :class:`~.Dict`.

    Attributes
    ---------
    type: :class:`~.Plane`
        Path to the dictionary of planes to be monitored.
    """
    @Dict.assign
    def __init__(self, values: dict):
        """The constructor for the :class:`~.Planes` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the planes to be monitored.

        """
        self.type = Plane


class FullDomain(Dict):
    """The :class:`~.FullDomain` class. It inherits from
    :class:`~.Dict`.

    Attributes
    ---------
    type: :obj:`bool`
        The boolean value to idicate whether to evaluate the domain.
    """
    @Dict.assign
    def __init__(self, value: bool):
        """The constructor for the :class:`~.FullDomain` class.

        Parameters
        ----------
        values : :obj:`dict`
            The values corresponding to the domain to be monitored.

        """
        self.type = bool 


class Plane(Settings):
    """The :class:`~.Plane` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    axis: :obj:`str`
        Either an xy, zy, or xz plane.

    distance: :class:`~.NonNegativeFloat`
        How far along the remaining axis
        this plane is.
    """
    @Settings.assign
    def __init__(self, value: dict):
        """The constructor for the :class:`~.Plane` class.

        Parameters
        ----------
        value : :obj:`dict`
            The values corresponding to each plane.
        """
        self.axis = Axis
        self.distance = NonNegativeFloat


class Axis(StringSelection):
    """The axis selection setting class.

    Attributes
    ----------
    options : :obj:`list` [:obj:`str`]
        The list of allowed options.

    """
    @Terminus.assign
    def __init__(self, value: str):
        """The constructor for the :class:`Axis` class.

        Parameters
        ----------
        value : :obj:`str`
            The chosen value.

        """
        self.options = [
            "xy",
            "yz",
            "xz"
        ]

    def check(self):
        pass


class ManualContours(Settings):
    """The :class:`~.ManualContours` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    min: :class:`~.NonNegativeFloat`
        The min value of the contours.
    max: :class:`~.NonNegativeFloat`
        The max value of the contours.

    """
    @Settings.assign
    def __init__(self, value: dict):
        """The constructor for the :class:`~.ManualContours` class.

        Parameters
        ----------
        value : :obj:`dict`
            The values corresponding to manual contours.
        """
        self.min = PositiveFloat 
        self.max = PositiveFloat


class WellMixed(Settings):
    """The :class:`~.WellMixed` class. It inherits from
    :class:`~.Settings`.

    Attributes
    ---------
    volume: :class:`~.NonNegativeFloat`
        The total volume of the container.
    """
    @Settings.assign
    def __init__(self, value: dict):
        """The constructor for the :class:`~.WellMixed` class.

        Parameters
        ----------
        value : :obj:`dict`
            The values corresponding to well mixed model.
        """
        self.volume = NonNegativeFloat


class Percentage(Number):
    """The :class:`~.Percentage` class. It inherits from
    :class:`~.Number`.

    Attributes
    ---------
    type :obj:`float`
        The float value corresponding to a percentage.
    """
    @Terminus.assign
    def __init__(self, value: float):
        """The constructor for the :class:`~.Percentage` class.

        Parameters
        ----------
        value : :obj:`float`
            The float value that is being checked.
        """
        self.type = float

    def check(self):
        """Abstract method from :class:`~.Terminus`.
        Raises
        ------
        ValueError
            If the value doesn't lie within the bounds.
        """
        self.lower_bound(0.0)
        self.upper_bound(100.0)


class NonNegativeFloat(Number):
    """The :class:`~.NonNegativeFloat` class. It inherits from
    :class:`~.Number`.

    """
    @Terminus.assign
    def __init__(self, value: float):
        """The constructor for the :class:`~.NonNegativeFloat` class.

        Parameters
        ----------
        value : :obj:`float`
            The float value that is being checked.
        """
        self.type = float

    def check(self):
        """Abstract method from :class:`~.Terminus`.
        Raises
        ------
        ValueError
            If the value isn't greater than 0.
        """
        self.lower_bound(0.0)


class PositiveFloat(Number):
    """The :class:`~.PositiveFloat` class. It inherits from
    :class:`~.Number`.

    """
    @Terminus.assign
    def __init__(self, value: float):
        """The constructor for the :class:`~.PositiveFloat` class.

        Parameters
        ----------
        value : :obj:`float`
            The float value that is being checked.
        """
        self.type = float

    def check(self):
        """Abstract method from :class:`~.Terminus`.
        Raises
        ------
        ValueError
            If the value isn't greater than 0.
        """
        self.lower_bound_exclusive(0.0)


class PositiveInt(Number):
    """The :class:`~.PositiveInt` class. It inherits from
    :class:`~.Number`.

    """
    @Terminus.assign
    def __init__(self, value: int):
        """The constructor for the :class:`~.PositiveInt` class.

        Parameters
        ----------
        value : :obj:`float`
            The float value that is being checked.
        """
        self.type = int

    def check(self):
        """Abstract method from :class:`~.Terminus`.
        Raises
        ------
        ValueError
            If the value isn't greater than 0.
        """
        self.lower_bound_exclusive(0)


class NonNegativeInteger(Number):
    """The :class:`~.NonNegativeInteger` class. It inherits from
    :class:`~.Number`.


    """
    @Terminus.assign
    def __init__(self, value: float):
        """The constructor for the :class:`~.NonNegativeInteger` class.

        Parameters
        ----------
        value : :obj:`float`
            The float value that is being checked.
        """
        self.type = int

    def check(self):
        """Abstract method from :class:`~.Terminus`.

        Raises
        ------
        ValueError
            If the value isn't greater than 0.
        """
        self.lower_bound(0)



import unittest
import os
import time

from ridt.data import DirectoryAgent


class TestDirectoryAgent(unittest.TestCase):

    """Unit tests for the :class:`~.TestDirectoryAgent` class."""

    def setUp(self) -> None:

        """setUp method which creates a dummy data :obj:`dict`
        which is used within the tests. Additionally it
        instantiates the :class:`~.DirectoryAgent` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        self.data = {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5}
        self.outdir = os.path.join(this_dir, "test_resources/directoryagent")
        self.run_index = 3
        self.shape = (5, 5)
        self.da = DirectoryAgent(self.outdir, self.shape)
